import datetime
from concurrent.futures import ThreadPoolExecutor
from unittest import TestCase, mock

from flask_jwt_extended import create_access_token

from dm.domain.entities import ActionTemplate, ActionType, Orchestration, Server
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.deployment import UndoCommand, CompositeCommand, CompletedProcess, Command, \
    create_cmd_from_orchestration, ProxyCommand, ProxyUndoCommand
from dm.utils.helpers import get_now
from dm.utils.var_context import VarContext
from dm.web import create_app, db


class TestCompositeCommand(TestCase):
    maxDiff = None

    def test_invoke_undo__stop_on_error_true(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=True, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertTrue(res)

    def test_invoke_undo__stop_on_error_false(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=False, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(5, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertTrue(res)

    def test_invoke_undo__step_stop_on_error_true(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, stop_on_error=True, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=False, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertTrue(res)

    def test_invoke_undo__stop_undo_on_error_true(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=False, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(4, mocked_imp_succ.execute.call_count)
        self.assertEqual(2, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

    def test_invoke_undo__step_stop_undo_on_error_true(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, stop_undo_on_error=True, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=True, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(2, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

    def test_invoke_undo__step_undo_on_error_false(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=get_now(),
                                                                end_time=get_now() + datetime.timedelta(
                                                                    5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=get_now(),
                                                                 end_time=get_now() + datetime.timedelta(
                                                                     5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_command=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_error, undo_command=uc2, undo_on_error=False, id_=2)
        c3 = Command(implementation=mocked_imp_succ, undo_command=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, stop_on_error=True, stop_undo_on_error=False, id_=1)

        res = cc1.invoke()

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertTrue(res)

    def test_result(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=start_time,
                                                                end_time=end_time)
        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=start_time,
                                                                 end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)

        ccu1 = CompositeCommand({uc1: []}, stop_on_error=True)
        ccu2 = CompositeCommand({uc2: [uc3], uc3: [uc4]}, stop_on_error=True)

        c1 = Command(implementation=mocked_imp_succ, undo_command=ccu1, id_=5)
        c2 = Command(implementation=mocked_imp_error, undo_command=ccu2, undo_on_error=False, id_=6)

        cc = CompositeCommand({c1: [c2], c2: []}, stop_on_error=True, stop_undo_on_error=False)

        res = cc.invoke()

        self.assertDictEqual({5: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              6: CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time)}
                             , cc.result)

        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertEqual(False, res)

        res = cc.undo()

        self.assertDictEqual({5: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              6: CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              1: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time)}
                             , cc.result)

        self.assertEqual(True, res)
        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

    def test_composite_command_success(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=start_time,
                                                                end_time=end_time)
        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=start_time,
                                                                 end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)

        ccu1 = CompositeCommand({uc1: []}, stop_on_error=True)
        ccu2 = CompositeCommand({uc2: [uc3], uc3: [uc4]}, stop_on_error=True)

        c1 = Command(implementation=mocked_imp_succ, undo_command=ccu1, id_=5)
        c2 = Command(implementation=mocked_imp_succ, undo_command=ccu2, id_=6)

        cc = CompositeCommand({c1: [c2], c2: []}, stop_on_error=False, stop_undo_on_error=False)

        res = cc.invoke()

        self.assertDictEqual({5: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              6: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time)}
                             , cc.result)
        self.assertEqual(True, res)
        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(0, mocked_imp_error.execute.call_count)

    def test_composite_command_error(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                                start_time=start_time,
                                                                end_time=end_time)
        mocked_imp_error.execute.return_value = CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                                 start_time=start_time,
                                                                 end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_error, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)
        uc5 = UndoCommand(implementation=mocked_imp_succ, id_=5)

        executor = ThreadPoolExecutor()
        ccu1 = CompositeCommand({uc1: []}, stop_on_error=True)
        ccu2 = CompositeCommand({uc2: [uc3, uc4]}, stop_on_error=True, executor=executor)

        c6 = Command(implementation=mocked_imp_succ, undo_command=ccu1, id_=6)
        c7 = Command(implementation=mocked_imp_succ, undo_command=ccu2, id_=7)
        c8 = Command(implementation=mocked_imp_succ, undo_command=uc5, id_=8)

        cc = CompositeCommand({c6: [c7], c7: [c8]}, stop_on_error=False, stop_undo_on_error=False, executor=executor)

        res = cc.invoke()

        self.assertTrue(res)
        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(0, mocked_imp_error.execute.call_count)

        res = cc.undo()

        self.assertFalse(res)
        self.assertEqual(7, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

        self.assertDictEqual({1: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              2: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              3: CompletedProcess(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              4: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              5: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              6: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              7: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time),
                              8: CompletedProcess(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                  start_time=start_time,
                                                  end_time=end_time)
                              }
                             , cc.result)


class TestCreateCmdFromOrchestration2(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    # @mock.patch('dm.use_cases.deployment.create_operation', autospec=IOperationEncapsulation)
    def test_create_cmd_from_orchestration(self):

        at = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0001', name='create dir', version=1,
                            action_type=ActionType.SHELL, code='mkdir {{dir}}',
                            parameters={}, expected_output='',
                            expected_rc=0, system_kwargs={})

        o = Orchestration('Test Orchestration', 1, 'description', id='bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

        me = Server('me', port=5000, me=True, id='cccccccc-1234-5678-1234-cccccccc0001')
        remote = Server('remote', port=5000, id='cccccccc-1234-5678-1234-cccccccc0002')

        db.session.add_all([me, remote, o])

        s1 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0001', undo=False, action_template=at,
                        parents=[], target=['frontend'])
        s2 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0002', undo=True, action_template=at,
                        parents=[s1], stop_on_error=False, target=[])
        s3 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0003', undo=False, action_template=at,
                        parents=[s1], stop_on_error=False, stop_undo_on_error=False, target=['frontend'])
        s4 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0004', undo=True, action_template=at,
                        parents=[s3], target=[])
        s5 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0005', undo=True, action_template=at,
                        parents=[s4], stop_on_error=True, target=[])
        s6 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0006', undo=True, action_template=at,
                        parents=[s4, s5], target=[])
        s7 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0007', undo=False, action_template=at,
                        parents=[s3], undo_on_error=False, target=[])
        s8 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0008', undo=True, action_template=at,
                        parents=[s7], target=[])
        s9 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0009', undo=False, action_template=at,
                        children=[s2, s3], target=['backend'])

        cc = create_cmd_from_orchestration(o, VarContext({'dir': 'C:\\test_folder'}),
                                            hosts={'all': [me.id, remote.id], 'frontend': [me.id],
                                                   'backend': [remote.id]},
                                            executor=None, register=mock.Mock())


        c1, c9 = cc._dag.get_nodes_at_level(1)
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0001'), c1.id)
        self.assertIsInstance(c1, Command)
        self.assertTrue(c1.stop_on_error)
        self.assertTrue(c1.undo_on_error)
        self.assertIsNone(c1.stop_undo_on_error)

        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0002',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0009'), c9.id)
        self.assertIsInstance(c9, Command)
        self.assertTrue(c9.stop_on_error)
        self.assertTrue(c9.undo_on_error)
        self.assertIsNone(c9.stop_undo_on_error)

        c21 = c1.undo_command
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0002'), c21.id)
        self.assertIsInstance(c21, UndoCommand)
        self.assertFalse(c21.stop_on_error)

        c22 = c9.undo_command
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0002',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0002'), c22.id)
        self.assertIsInstance(c22, UndoCommand)
        self.assertFalse(c22.stop_on_error)

        c3, = cc._dag.get_nodes_at_level(2)
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0003'), c3.id)
        self.assertIsInstance(c3, Command)
        self.assertFalse(c3.stop_on_error)
        self.assertTrue(c3.undo_on_error)
        self.assertFalse(c3.stop_undo_on_error)

        self.assertTupleEqual(('undo',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0003'), c3.undo_command.id)
        self.assertIsInstance(c3.undo_command, CompositeCommand)
        self.assertFalse(c3.undo_command.stop_on_error)
        self.assertIsNone(c3.undo_command.stop_undo_on_error)

        c4, = c3.undo_command._dag.get_nodes_at_level(1)
        self.assertIsInstance(c4, UndoCommand)
        self.assertFalse(c4.stop_on_error)

        c5, = c3.undo_command._dag.get_nodes_at_level(2)
        self.assertIsInstance(c5, UndoCommand)
        self.assertTrue(c5.stop_on_error)

        c6, = c3.undo_command._dag.get_nodes_at_level(3)
        self.assertIsInstance(c6, UndoCommand)
        self.assertFalse(c6.stop_on_error)

        cc7, = cc._dag.get_nodes_at_level(3)
        self.assertEqual('eeeeeeee-1234-5678-1234-eeeeeeee0007', cc7.id)
        self.assertIsInstance(cc7, CompositeCommand)
        self.assertFalse(cc7.stop_on_error)
        self.assertFalse(cc7.stop_undo_on_error)

        c71, c72 = cc7._dag.root
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0007'), c71.id)
        self.assertIsInstance(c71, Command)
        self.assertTrue(c71.stop_on_error)
        self.assertFalse(c71.undo_on_error)
        self.assertIsNone(c71.stop_undo_on_error)

        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0008'), c71.undo_command.id)
        self.assertIsInstance(c71.undo_command, UndoCommand)
        self.assertTrue(c71.undo_command.stop_on_error)

        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0002',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0007'), c72.id)
        self.assertIsInstance(c72, ProxyCommand)
        self.assertTrue(c72.stop_on_error)
        self.assertFalse(c72.undo_on_error)
        self.assertIsNone(c72.stop_undo_on_error)

        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0002',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0008'), c72.undo_command.id)
        self.assertIsInstance(c72.undo_command, ProxyUndoCommand)
        self.assertTrue(c72.undo_command.stop_on_error)

    def test_create_cmd_from_orchestration_one_step(self):
        at = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0001', name='create dir', version=1,
                            action_type=ActionType.SHELL, code='mkdir {{dir}}',
                            parameters={}, expected_output='',
                            expected_rc=0, system_kwargs={})

        o = Orchestration('Test Orchestration', 1, 'description',
                          id='bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

        me = Server('me', port=5000, me=True, id='cccccccc-1234-5678-1234-cccccccc0001')
        remote = Server('remote', port=5000, id='cccccccc-1234-5678-1234-cccccccc0002')

        db.session.add_all([me, remote, o])

        s1 = o.add_step(id='eeeeeeee-1234-5678-1234-eeeeeeee0001', undo=False, action_template=at,
                        parents=[], target=[])

        cc = create_cmd_from_orchestration(o, VarContext({'dir': './test_folder'}),
                                            hosts={'all': [me.id]},
                                            executor=None, register=mock.Mock())

        c1,  = cc._dag.get_nodes_at_level(1)
        self.assertTupleEqual(('cccccccc-1234-5678-1234-cccccccc0001',
                               'eeeeeeee-1234-5678-1234-eeeeeeee0001'), c1.id)
        self.assertIsInstance(c1, Command)
        self.assertTrue(c1.stop_on_error)
        self.assertTrue(c1.undo_on_error)
        self.assertIsNone(c1.stop_undo_on_error)



import hashlib
import os
from unittest import mock, TestCase

from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, Software, SoftwareServerAssociation
from dm.domain.entities import User
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db, errors
from dm.web.network import Response
from tests.helpers import ValidateResponseMixin


class Test(TestCase, ValidateResponseMixin):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

        self.source_path = '/software'
        self.filename = 'filename.zip'
        self.content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        self.size = len(self.content)
        self.checksum = hashlib.md5(self.content).hexdigest()
        self.dest_path = '/dest_repo'

        self.soft = Software(name='test_software', version=1, filename=self.filename, size=self.size,
                             checksum=self.checksum)
        self.ssa = SoftwareServerAssociation(software=self.soft, server=Server.get_current(), path=self.source_path)
        self.node2 = Server('node2', port=5000)
        db.session.add_all([self.soft, self.ssa, self.node2])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_software(self, mock_post, mock_send_file, mock_exists):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = True

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path),
                                headers=self.auth.header)

        mock_post.assert_called_once()
        server, view = mock_post.call_args[0]
        kwargs = mock_post.call_args[1]
        self.assertEqual(self.node2, db.session.merge(server))
        self.assertDictEqual({'software_id': str(self.soft.id), 'num_chunks': 1,
                              'dest_path': self.dest_path}, kwargs['json'])
        self.assertEqual(202, resp.status_code)
        self.assertDictEqual({'transfer_id': '1'}, resp.get_json())

    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_foreground(self, mock_post, mock_send_file, mock_exists):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = True

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path, background=False),
                                headers=self.auth.header)

        mock_post.assert_called_once()
        server, view = mock_post.call_args[0]
        kwargs = mock_post.call_args[1]
        self.assertEqual(self.node2, db.session.merge(server))
        self.assertDictEqual({'software_id': str(self.soft.id), 'num_chunks': 1,
                              'dest_path': self.dest_path}, kwargs['json'])
        self.assertEqual(201, resp.status_code)
        self.assertDictEqual({'transfer_id': '1'}, resp.get_json())

    @mock.patch('dm.web.api_1_0.urls.use_cases.md5', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.getsize', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_file(self, mock_post, mock_send_file, mock_exists, mock_getsize, mock_md5):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = True
        mock_getsize.return_value = self.size
        mock_md5.return_value = self.checksum

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(file=os.path.join(self.source_path, self.filename),
                                          dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path),
                                headers=self.auth.header)

        mock_post.assert_called_once()
        server, view = mock_post.call_args[0]
        kwargs = mock_post.call_args[1]
        self.assertEqual(self.node2, db.session.merge(server))
        self.assertDictEqual({'filename': self.filename, 'num_chunks': 1, 'dest_path': self.dest_path,
                              'checksum': self.checksum, 'size': self.size},
                             kwargs['json'])
        self.assertEqual(202, resp.status_code)
        self.assertDictEqual({'transfer_id': '1'}, resp.get_json())

    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_SoftwareServerNotFound(self, mock_post, mock_send_file, mock_exists):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = True
        db.session.delete(self.ssa)
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path),
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.SoftwareServerNotFound(str(self.soft.id), str(self.node2.id)))

    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_software_FileNotFound(self, mock_post, mock_send_file, mock_exists):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = False

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path),
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.FileNotFound(os.path.join(self.source_path, self.filename)))

    @mock.patch('dm.web.api_1_0.urls.use_cases.md5', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.getsize', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_file_FileNotFound(self, mock_post, mock_send_file, mock_exists, mock_getsize, mock_md5):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = False
        mock_getsize.return_value = self.size
        mock_md5.return_value = self.checksum

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(file=os.path.join(self.source_path, self.filename),
                                          dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path),
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.FileNotFound(os.path.join(self.source_path, self.filename)))

    @mock.patch('dm.web.api_1_0.urls.use_cases.get', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.os.path.exists', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.async_send_file', autospec=True)
    @mock.patch('dm.web.api_1_0.urls.use_cases.post', autospec=True)
    def test_send_software_get_transfer_data(self, mock_post, mock_send_file, mock_exists, mock_get):
        mock_post.return_value = Response(msg={'transfer_id': '1'}, code=200)
        mock_exists.return_value = True
        mock_get.return_value = Response(msg={'transfer_id': '1', 'status': 'COMPLETED'}, code=200)

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path, include_transfer_data=True),
                                headers=self.auth.header)

        mock_get.assert_called_once()
        server, view = mock_get.call_args[0]
        kwargs = mock_get.call_args[1]
        self.assertEqual(self.node2, db.session.merge(server))
        self.assertEqual('api_1_0.transferresource', view)
        self.assertDictEqual({'transfer_id': '1'}, kwargs['view_data'])

        self.assertEqual(202, resp.status_code)
        self.assertDictEqual({'transfer_id': '1', 'status': 'COMPLETED'}, resp.get_json())

        mock_get.return_value = Response(msg={'error': {'message': 'some error content'}}, code=404)

        resp = self.client.post(url_for('api_1_0.send'),
                                json=dict(software_id=str(self.soft.id), dest_server_id=str(self.node2.id),
                                          dest_path=self.dest_path, include_transfer_data=True),
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.HTTPError(mock_get.return_value))


"""
Utility function and classes for dm

"""
import copy
import typing as t


class DAGError(Exception):
    """Base Class Exception in DAG"""


T = t.TypeVar('T')
Tree = t.Dict[T, t.List[T]]


class DAG(t.Generic[T]):
    """
    Base class for directed graphs.
    A DAG stores nodes and edges
    DAG hold directed edges.  Self loops are allowed but multiple
    (parallel) edges are not.
    Nodes can be arbitrary (hashable) Python objects.
    By convention `None` is not used as a node.
    Edges are represented as links between nodes.
    ----------
    Create an empty graph structure (a "null graph") with no nodes and
    no edges.
    >>> G = DAG()
    G can be grown in several ways.
    **Nodes:**
    Add one node at a time:
    >>> G.add_node(1)
    Add the nodes from any container (a list, dict, set or
    even the lines from a file or the nodes from another graph).
    >>> G.add_nodes_from([2, 3])
    >>> G.add_nodes_from(range(100, 110))
    >>> H = DAG([(2,3),(3,4)])
    >>> G.add_nodes_from(H)
    In addition to strings and integers any hashable Python object
    (except None) can represent a node, e.g. a customized node object,
    or even another Graph.
    >>> G.add_node(H)
    **Edges:**
    G can also be grown by adding edges.
    Add one edge,
    >>> G.add_edge(1, 2)
    a list of edges,
    >>> G.add_edges_from([(1, 2), (1, 3)])
    or a collection of edges,
    >>> G.add_edges_from(H.edges)
    If some edges connect nodes not yet in the graph, the nodes
    are added automatically.  There are no errors when adding
    nodes or edges that already exist.
    Many common graph features allow python syntax to speed reporting.
    >>> 1 in G     # check if node in graph
    True
    >>> [n for n in G if n < 3]  # iterate through nodes
    [1, 2]
    >>> len(G)  # number of nodes in graph
    5

    """

    def __init__(self, edges_for_adding: t.List[t.Tuple[T, T]] = None):
        self._nodes = []
        self._pred = {}
        self._succ = {}
        if edges_for_adding:
            self.add_edges_from(edges_for_adding)

    @property
    def nodes(self) -> t.List[T]:
        return self._nodes

    @property
    def ordered_nodes(self) -> t.List[T]:
        nodes = []
        for l in range(1, self.depth + 1):
            nodes.extend(self.get_nodes_at_level(l))
        return nodes

    @property
    def pred(self) -> t.Dict[T, t.List[T]]:
        return self._pred

    @property
    def succ(self) -> t.Dict[T, t.List[T]]:
        return self._succ

    @property
    def root(self) -> t.List[T]:
        return [k for k, v in self._pred.items() if len(v) == 0]

    def __iter__(self) -> t.Iterator[T]:
        """Iterate over the nodes. Use: 'for n in G'.
        Returns
        -------
        niter : iterator
            An iterator over all nodes in the graph.
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> [n for n in G]
        [0, 1, 2, 3]
        >>> list(G)
        [0, 1, 2, 3]
        """
        return iter(self._nodes)

    def __contains__(self, n) -> bool:
        """Returns True if n is a node, False otherwise. Use: 'n in G'.
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> 1 in G
        True
        """
        try:
            return n in self._nodes
        except TypeError:
            return False

    def __len__(self) -> int:
        """Returns the number of nodes. Use: 'len(G)'.
        Returns
        -------
        nodes:
            The number of nodes in the graph.
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> len(G)
        4
        """
        return len(self._nodes)

    def __getitem__(self, n: T) -> t.List[T]:
        """Returns a dict of neighbors of node n.  Use: 'G[n]'.
        Parameters
        ----------
        n:
           A node in the graph.
        Returns
        -------
        adj_list:
           The successor list for nodes connected to n.
        Notes
        -----
        G[n] is the same as G.adj[n] and similar to G.neighbors(n)
        (which is an iterator over G.adj[n])
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> G[1]
        [2]
        """
        return self._succ[n]

    def add_node(self, node_for_adding: T) -> 'DAG':
        """Returns the graph with the added node
        Parameters
        ----------
        node_for_adding : node
           A node in the graph.
        Returns
        -------
        graph : DAG
          The graph with the node added
        Examples
        --------
        >>> G = DAG()
        >>> G.add_node(1)
        >>> G.nodes
        [1]
        """
        if node_for_adding not in self._nodes:
            self._succ[node_for_adding] = []
            self._pred[node_for_adding] = []
            self._nodes.append(node_for_adding)

    def add_nodes_from(self, nodes_for_adding: t.List[T]):
        for n in nodes_for_adding:
            self.add_node(n)

    def remove_node(self, n: T) -> None:
        """Remove node n.
        Removes the node n and all adjacent edges.
        Attempting to remove a non-existent node will raise an exception.
        Parameters
        ----------
        n:
           A node in the graph
        Raises
        -------
        NetworkXError
           If n is not in the graph.
        See Also
        --------
        remove_nodes_from
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> list(G.edges)
        [(1, 2), (2, 3), (3, 4)]
        >>> G.remove_node(2)
        >>> G.edges
        [(3, 4)]
        """
        try:
            nbrs = self._succ[n]

        except KeyError:  # NetworkXError if n not in self
            raise DAGError("The node %s is not in the digraph." % (n,))
        for u in nbrs:
            self._pred[u].remove(n)
        for u in self._pred[n]:
            self._succ[u].remove(n)  # remove all edges n-u in digraph
        self._nodes.remove(n)
        del self._pred[n]
        del self._succ[n]

    def remove_nodes_from(self, nodes: t.Iterable[T]):
        """Remove multiple nodes.
        Parameters
        ----------
        nodes : iterable container
            A container of nodes (list, dict, set, etc.).  If a node
            in the container is not in the graph it is silently ignored.
        See Also
        --------
        remove_node
        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> e = list(G.nodes)
        >>> e
        [0, 1, 2]
        >>> G.remove_nodes_from(e)
        >>> list(G.nodes)
        []
        """
        for n in nodes:
            try:
                self.remove_node(n)
            except DAGError:
                pass  # silent failure on remove

    def add_edge(self, u_of_edge, v_of_edge):
        """Add an edge between u and v.

        The nodes u and v will be automatically added if they are
        not already in the graph.

        Edge attributes can be specified with keywords or by directly
        accessing the edge's attribute dictionary. See examples below.

        Parameters
        ----------
        u, v : nodes
           Nodes can be, for example, strings or numbers.
           Nodes must be hashable (and not None) Python objects.

        See Also
        --------
        add_edges_from : add a collection of edges

        Notes
        -----
        Adding an edge that already exists doesn't do anything.

        Examples
        --------
        The following all add the edge e=(1, 2) to graph G:
        >>> G = DAG()
        >>> e = (1, 2)
        >>> G.add_edge(1, 2)           # explicit two-node form
        >>> G.add_edge(*e)             # single edge as tuple of two nodes
        >>> G.add_edges_from( [(1, 2)] ) # add edges from iterable container
        """

        u, v = u_of_edge, v_of_edge
        # add nodes
        if not u in self._nodes:
            self._nodes.append(u)
            self._succ[u] = []
            self._pred[u] = []
        if not v in self._nodes:
            self._nodes.append(v)
            self._succ[v] = []
            self._pred[v] = []
        if v not in self._succ[u]:
            self._succ[u].append(v)
            self._pred[v].append(u)

    def add_edges_from(self, ebunch_to_add):
        """Add all the edges in ebunch_to_add.

        Parameters
        ----------
        ebunch_to_add : container of edges
            Each edge given in the container will be added to the
            graph. The edges must be given as 2-tuples (u, v).


        See Also
        --------
        add_edge : add a single edge

        Notes
        -----
        Adding the same edge twice has no effect.

        Examples
        --------
        >>> G = DAG()
        >>> G.add_edges_from([(0, 1), (1, 2)]) # using a list of edge tuples
        >>> e = zip(range(0, 3), range(1, 4))
        >>> G.add_edges_from(e) # Add the path graph 0-1-2-3
        """
        for e in ebunch_to_add:
            ne = len(e)
            if ne == 2:
                u, v = e
            else:
                raise DAGError(
                    "Edge tuple %s must be a 2-tuple" % (e,))
            self.add_edge(u, v)

    def remove_edge(self, u, v):
        """Remove the edge between u and v.

        Parameters
        ----------
        u, v : nodes
            Remove the edge between nodes u and v.

        Raises
        ------
        DAGError
            If there is not an edge between u and v.

        See Also
        --------
        remove_edges_from : remove a collection of edges

        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> G.remove_edge(2, 3)
        >>> e = (1, 2)
        >>> G.remove_edge(*e) # unpacks e from an edge tuple
        """
        try:
            self._succ[u].remove(v)
            self._pred[v].remove(u)
        except KeyError:
            raise DAGError("The edge %s-%s not in graph." % (u, v))

    def remove_edges_from(self, ebunch):
        """Remove all edges specified in ebunch.

        Parameters
        ----------
        ebunch: list or container of edge tuples
            Each edge given in the list or container will be removed
            from the graph. The edges must be  2-tuples (u, v) edge between u and v.

        See Also
        --------
        remove_edge : remove a single edge

        Notes
        -----
        Will fail silently if an edge in ebunch is not in the graph.

        Examples
        --------
        >>> G = DAG([(1,2),(2,3),(3,4)])
        >>> ebunch = [(1, 2), (2, 3)]
        >>> G.remove_edges_from(ebunch)
        """
        for e in ebunch:
            u, v = e[:2]  # ignore edge data
            if u in self._succ and v in self._succ[u]:
                self._succ[u].remove(v)
                self._pred[v].remove(u)

    def level(self, node):
        """The level of a node is one greater than the level of its parent.
        The level of the root node is 1.

        Parameters
        ----------
        node: Any
            node from which we want to get it's level

        Examples
        --------
        >>> G = DAG([('a','b'),('b','c'),('c','d')])
        >>> G.level('c')
        3
        >>> G.add_edges_from([('a','e'), ('e','b')]
        >>> G.level('c')
        4

        Returns
        -------
        int:
            level of corresponding node
        """
        if len(self._pred[node]) == 0:
            return 1
        else:
            levels = [self.level(s) for s in self._pred[node]]
            if levels:
                return max(levels) + 1
            else:
                return 1

    # def get_nodes_at_level(self, level, step=None, nodes=None, visited_nodes=None):
    #     """Returns a list with a nodes at level
    #
    #      Parameters
    #     ----------
    #     level: level from which you want to get the nodes
    #
    #     Examples
    #     --------
    #     >>> G = DAG([('a','b'),('b','c'),('c','d')])
    #     >>> G.get_nodes_at_level(3)
    #     ['c']
    #     >>> G.add_edges_from([('b','e')]
    #     >>> G.get_nodes_at_level(3)
    #     ['c', 'e']
    #     """
    #     nodes = nodes or []
    #     visited_nodes = visited_nodes or []
    #
    #     if step == None:
    #         if self.root_step.height == height:
    #             nodes.append(self.root_step)
    #             return nodes
    #         else:
    #             for s in self.root_step.child_do_steps:
    #                 if not s in visited_nodes:
    #                     self._get_steps_at_height(height, s, nodes, visited_nodes)
    #                     visited_nodes.append(s)
    #             return nodes
    #     else:
    #         if step.height == height:
    #             nodes.append(step)
    #         elif step.height > height:
    #             for s in step.child_do_steps:
    #                 if not s in visited_nodes:
    #                     self._get_steps_at_height(height, s, nodes, visited_nodes)
    #                     visited_nodes.append(s)

    def _is_cyclic_util(self, v, visited, rec_stack):
        """Depth First Traversal can be used to detect a cycle in a Graph.
        DFS for a connected graph produces a tree. There is a cycle in a graph only if there is
        a back edge present in the graph. A back edge is an edge that is from a node to
        itself (self-loop) or one of its ancestor in the tree produced by DFS.

        To detect a back edge, we can keep track of vertices currently in recursion stack of
        function for DFS traversal. If we reach a vertex that is already in the recursion stack,
        then there is a cycle in the tree. The edge that connects current vertex to the vertex in
        the recursion stack is a back edge. We have used recStack[] array to keep track of vertices
        in the recursion stack.

        Parameters
        ----------
        v: vertex to make the DFS
        visited: dict with all nodes. A nodes visited has true inside
        rec_stack: to keep track of vertices in the recursion stack

        Reference
        ---------
        https://www.geeksforgeeks.org/detect-cycle-in-a-graph/
        """

        # Mark current node as visited and
        # adds to recursion stack
        visited[v] = True
        rec_stack[v] = True

        # Recur for all neighbours
        # if any neighbour is visited and in
        # recStack then graph is cyclic
        for neighbour in self._succ[v]:
            if not visited[neighbour]:
                if self._is_cyclic_util(neighbour, visited, rec_stack):
                    return True
            elif rec_stack[neighbour]:
                return True

        # The node needs to be poped from
        # recursion stack before function ends
        rec_stack[v] = False
        return False

    # Returns true if graph is cyclic else false
    def is_cyclic(self):
        """
        Given a directed graph, check whether the graph contains a cycle or not.
        Function returns true if the given graph contains at least one cycle, else returns false

        Return
        ------
        boolean

        """
        visited = {n: False for n in self._nodes}
        rec_stack = {n: False for n in self._nodes}
        for node in self._nodes:
            if visited[node] == False:
                if self._is_cyclic_util(node, visited, rec_stack):
                    return True
        return False

    def to_dict_of_lists(self):
        """Converts the tree into a dict of lists containing every node and its successors

        Examples
        --------
        >>> G = DAG([(1, 2),(2, 3),(2, 4)])
        >>> G.to_dict_of_lists()
        {1:[2], 2: [3, 4], 3: [], 4:[]}
        """

        return self._succ

    @classmethod
    def from_dict_of_lists(cls, dtree):
        g = cls()
        g.add_nodes_from((n for n in dtree))

        for k, v in dtree.items():
            for n in v:
                g.add_edge(k, n)

        return g

    def get_nodes_at_level(self, level):
        """

        Parameters
        ----------
        level

        Returns
        -------

        """
        return [n for n in self if self.level(n) == level]

    @property
    def depth(self):
        return max([self.level(n) for n in self])

    def copy(self) -> 'DAG':
        """
        Makes a copy of the current DAG. All the node objects remain the same. We make a copy of the graph dependency

        Returns
        -------
        dag:
            graph copied
        """
        g = DAG()
        g._nodes = copy.copy(self._nodes)
        g._succ = {k: copy.copy(v) for k, v in self._succ.items()}
        g._pred = {k: copy.copy(v) for k, v in self._pred.items()}
        return g

    def _fill_subtree(self, node: T, subtree: Tree):
        for n in self.succ[node]:
            if n not in subtree:
                self._fill_subtree(n, subtree)
        subtree.update({node: self.succ[node]})

    def subtree(self, starts_with: t.Union[t.Iterable[T], t.List[T]]) -> Tree:
        st = {}
        for n in starts_with:
            if n not in st:
                self._fill_subtree(n, st)
        return st



import time
from unittest import TestCase, mock
from unittest.mock import patch

from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Orchestration, ActionTemplate, ActionType, Server, User
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db, errors
from tests.helpers import ValidateResponseMixin


class TestLaunchOrchestration(TestCase, ValidateResponseMixin):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial()
        self.u = User('test')
        db.session.add(self.u)
        db.session.commit()
        self.auth = HTTPBearerAuth(create_access_token(self.u.id))
        self.o = Orchestration('create user', version=1)
        self.at1 = ActionTemplate(action_type=ActionType.SHELL, name="create folder", version=1,
                                  code="sudo mkdir -p {{folder}}")
        self.at2 = ActionTemplate(action_type=ActionType.SHELL, name="delete folder", version=1,
                                  code="sudo rm -fr {{folder}}")
        self.at3 = ActionTemplate(action_type=ActionType.SHELL, name="create user", version=1,
                                  code="sudo useradd -d {{home}} {{user}}")
        self.at4 = ActionTemplate(action_type=ActionType.SHELL, name="delete user", version=1,
                                  code="sudo userdel {{user}}")

        self.s1 = self.o.add_step(undo=False, action_template=self.at1)
        self.s2 = self.o.add_step(undo=True, action_template=self.at2, parents=[self.s1])
        self.s3 = self.o.add_step(undo=False, action_template=self.at3, parents=[self.s1])
        self.s4 = self.o.add_step(undo=True, action_template=self.at4, parents=[self.s3])
        db.session.add(self.o)
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_launch_orchestration_error_on_server(self):
        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=str(self.o.id)),
                                json={'hosts': 'aaaaaaaa-1234-5678-1234-56781234aaa1'},
                                headers=self.auth.header)

        self.assertEqual(404, resp.status_code)
        self.assertDictEqual(
            errors.format_error_content(errors.ServerNormalizationError(['aaaaaaaa-1234-5678-1234-56781234aaa1'])),
            resp.get_json())

    def test_launch_orchestration_error_on_target(self):
        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=str(self.o.id)),
                                json={'hosts': {'all': ['aaaaaaaa-1234-5678-1234-56781234aaa1'],
                                                'other': 'granule'}},
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.TargetNotNeeded(['other']))

        s = self.o.steps[0]
        s.target = ['new']
        db.session.commit()

        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=str(self.o.id)),
                                json={'hosts': {'all': ['aaaaaaaa-1234-5678-1234-56781234aaa1']}},
                                headers=self.auth.header)

        self.validate_error_response(resp, errors.TargetUnspecified(['new']))

    @patch('dm.web.api_1_0.urls.use_cases.uuid.uuid4', return_value='a7083c43-34cc-4b26-91f0-ea0928cf5945')
    @patch('dm.web.api_1_0.urls.use_cases.VarContext')
    @patch('dm.web.api_1_0.urls.use_cases.deploy_orchestration')
    @patch('dm.web.api_1_0.urls.use_cases.HTTPBearerAuth', return_value='token')
    def test_launch_orchestration_ok(self, mock_create, mock_deploy, mock_var_context, mock_uuid4):
        mock_uuid4.return_value = 1
        mock_create.return_value = 'token'
        data = {'hosts': str(Server.get_current().id),
                'params': {'folder': '/home/{{user}}',
                           'home': '{{folder}}',
                           'user': 'dimensigon'}}

        def deploy_orchestration(*args, **kwargs):
            return {'result': 'ok'}

        def deploy_orchestration_delayed(*args, **kwargs):
            time.sleep(1.1)
            return dict(**kwargs)

        mock_deploy.side_effect = deploy_orchestration
        MockVarContext = mock.Mock(name='MockVarContext')
        mock_var_context.return_value = MockVarContext
        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=str(self.o.id)),
                                json=data,
                                headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual({'result': 'ok'}, resp.get_json())

        mock_var_context.assert_called_once_with(
            globals=dict(execution_id='1', executor_id=str(self.u.id)), initials=data['params'])
        mock_deploy.assert_called_once_with(orchestration=self.o.id, var_context=MockVarContext,
                                            hosts={'all': [str(Server.get_current().id)]})

        mock_deploy.side_effect = deploy_orchestration_delayed
        resp = self.client.post(url_for('api_1_0.launch_orchestration', orchestration_id=str(self.o.id)),
                                json=data,
                                headers=self.auth.header)

        self.assertEqual(202, resp.status_code)
        self.assertDictEqual({'execution_id': '1'},
                             resp.get_json())



from enum import Enum

from dm.utils import typos
from dm.utils.typos import Pickle
from dm.web import db


class Scope(Enum):
    CATALOG = 1  # lock catalog for updating information
    ORCHESTRATION = 2  # execute an orchestration
    UPGRADE = 3  # upgrading Catalog from another server

    def __lt__(self, other):
        return self.value < other.value


class State(Enum):
    UNLOCKED = 1
    PREVENTING = 2
    LOCKED = 3


class Locker(db.Model):
    __tablename__ = 'L_locker'

    scope = db.Column(typos.Enum(Scope, name=True), primary_key=True)
    state = db.Column(typos.Enum(State), nullable=False)
    applicant = db.Column(Pickle)

    @classmethod
    def set_initial(cls):
        for scope in Scope:
            l = cls.query.get(scope)
            if not l:
                l = cls(scope=scope, state=State.UNLOCKED)
                db.session.add(l)
            else:
                if l.state != State.UNLOCKED:
                    l.state = State.UNLOCKED

    def __repr__(self):
        return f"Locker({self.scope.name}, {self.state.name})"

    def __str__(self):
        return f"Locker({self.scope.name}, {self.state.name})"



from .action_template import ActionTemplateList, ActionTemplateResource
from .execution import OrchExecutionList, OrchExecutionResource, OrchestrationExecutionRelationship, StepExecutionList, \
    StepExecutionResource, OrchExecStepExecRelationship
from .granule import GranuleList
from .log import LogList, LogResource
from .orchestration import OrchestrationList, OrchestrationResource
from .server import ServerList, ServerResource
from .software import SoftwareList, SoftwareResource, SoftwareServersResource
from .step import StepList, StepResource
from .step_children import StepRelationshipChildren
from .step_parents import StepRelationshipParents
from .transfer import TransferList, TransferResource
from .user import UserList, UserResource



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import StepExecution, OrchExecution
from dm.web.decorators import securizer, forward_or_dispatch
from dm.web.helpers import filter_query, check_param_in_uri


class StepExecutionList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(StepExecution, request.args)
        return [at.to_json(human=check_param_in_uri('human')) for at in query.order_by(StepExecution.start_time).all()]


class StepExecutionResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, execution_id):
        return StepExecution.query.get_or_404(execution_id).to_json(human=check_param_in_uri('human'))


class OrchestrationExecutionRelationship(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, orchestration_id):
        query = filter_query(OrchExecution, request.args)
        return [oe.to_json(human=check_param_in_uri('human')) for oe in
                query.filter_by(orchestration_id=orchestration_id).order_by(OrchExecution.start_time).all()]


class OrchExecStepExecRelationship(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, execution_id):
        query = filter_query(StepExecution, request.args)
        return [oe.to_json(human=check_param_in_uri('human')) for oe in
                query.filter_by(orch_execution_id=execution_id).order_by(StepExecution.start_time).all()]


class OrchExecutionList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(OrchExecution, request.args)
        return [oe.to_json(human=check_param_in_uri('human'), add_step_exec=check_param_in_uri('steps')) for oe in
                query.order_by(OrchExecution.start_time).all()]


class OrchExecutionResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, execution_id):
        return OrchExecution.query.get_or_404(execution_id).to_json(add_step_exec=check_param_in_uri('steps'),
                                                                    human=check_param_in_uri('human'))



import base64
import math
import os
import typing as t

import aiohttp
from flask import current_app, g

import dm.use_cases.lock as lock
from dm import defaults
from dm.domain.entities import Orchestration, Server, Scope, OrchExecution, User
from dm.use_cases.deployment import create_cmd_from_orchestration, RegisterStepExecution
from dm.utils import asyncio
from dm.utils.helpers import get_now
from dm.utils.typos import Id
from dm.web import db, errors, executor
from dm.web.network import async_post, async_put

if t.TYPE_CHECKING:
    from dm.utils.var_context import VarContext


async def async_send_file(dest_server: Server, transfer_id: Id, file,
                          chunk_size: int = None, chunks: int = None, max_senders: int = None,
                          auth=None, retries: int = 1):
    async def send_chunk(server: Server, view: str, chunk, sem):
        async with sem:
            json_msg = dict(chunk=chunk)

            with open(file, 'rb') as fd:
                fd.seek(chunk * chunk_size)
                chunk_content = base64.b64encode(fd.read(chunk_size)).decode('ascii')
            json_msg.update(content=chunk_content)

            return await async_post(server, view_or_url=view,
                                    view_data=dict(transfer_id=str(transfer_id)), json=json_msg, auth=auth,
                                    session=session)

    chunk_size = chunk_size or defaults.CHUNK_SIZE
    max_senders = max_senders or defaults.MAX_SENDERS
    chunks = chunks or math.ceil(os.path.getsize(file) / chunk_size)
    retries = retries
    sem = asyncio.Semaphore(max_senders)
    responses = {}
    l_chunks = [c for c in range(0, chunks)]
    async with aiohttp.ClientSession() as session:
        while retries > 0:
            retry_chunks = []
            for chunk in l_chunks:
                task = asyncio.create_task(send_chunk(dest_server, 'api_1_0.transferresource', chunk, sem))
                responses.update({chunk: task})

            for chunk, task in responses.items():
                resp = await task
                if resp.code != 201:
                    retry_chunks.append(chunk)
                elif resp.code == 410:
                    raise errors.TransferNotInValidState(transfer_id, resp.msg['error'].get(['status'], None))

            if len(retry_chunks) == 0 and chunks != 1:
                resp = await async_put(dest_server, 'api_1_0.transferresource',
                                       view_data={'transfer_id': transfer_id},
                                       session=session,
                                       auth=auth)
                if resp.code != 201:
                    current_app.logger.error(
                        f"Transfer {transfer_id}: Unable to create file at destination {dest_server.name}: "
                        f"{resp}")
            l_chunks = retry_chunks
            retries -= 1
    if l_chunks:
        data = {c: responses[c].result() for c in l_chunks}
        raise errors.ChunkSendError(data)


def deploy_orchestration(orchestration: t.Union[Id, Orchestration],
                         hosts: t.Dict[str, t.Union[t.List[Id]]],
                         var_context: 'VarContext' = None,
                         execution: t.Union[Id, OrchExecution] = None,
                         executor: t.Union[Id, User] = None,
                         execution_server: t.Union[Id, Server] = None) -> OrchExecution:
    """deploy the orchestration

    Args:
        orchestration: id or orchestration to execute
        hosts: Mapping to all distributions
        params: VarContext configuration
        execution: id or execution to associate with the orchestration. If none, a new one is created
        executor: id or User who executes the orchestration
        execution_server: id or User who executes the orchestration

    Returns:
        dict: dict with tuple ids and the :class: CompletedProcess

    Raises:
        Exception: if anything goes wrong
    """
    execution = execution or var_context.globals.get('execution_id')
    executor = executor or var_context.globals.get('executor_id')
    hosts = hosts or var_context.globals.get('hosts')
    if not isinstance(orchestration, Orchestration):
        orchestration = Orchestration.query.get(orchestration)
    if not isinstance(execution, OrchExecution):

        if execution is not None:
            exe = OrchExecution.query.get(execution)
        else:
            exe = OrchExecution.query.get(var_context.globals.get('execution_id'))
        if exe is None:
            if not isinstance(executor, User):
                executor = User.query.get(executor)
            if executor is None:
                raise ValueError('executor must be set')
            if not isinstance(execution_server, Server):
                if execution_server is None:
                    try:
                        execution_server = g.server
                    except AttributeError:
                        execution_server = Server.get_current()
                    if execution_server is None:
                        raise ValueError('execution server not found')
                else:
                    execution_server = Server.query.get(execution_server)
            exe = OrchExecution(id=execution, orchestration=orchestration, target=hosts, params=dict(var_context),
                                _executor=executor, server=execution_server)
            db.session.add(exe)
            db.session.commit()
    else:
        exe = execution
    current_app.logger.debug(
        f"Execution {exe.id}: Launching orchestration {orchestration} on {hosts} with {var_context}")
    return _deploy_orchestration(orchestration, var_context, hosts, exe)


def _deploy_orchestration(orchestration: Orchestration,
                          var_context: 'VarContext',
                          hosts: t.Dict[str, t.List[Id]],
                          execution: OrchExecution
                          ) -> OrchExecution:
    """
    Parameters
    ----------
    orchestration
        orchestration to deploy
    params
        parameters to pass to the steps

    Returns
    -------
    t.Tuple[bool, bool, t.Dict[int, dpl.CompletedProcess]]:
        tuple with 3 values. (boolean indicating if invoke process ended up successfully,
        boolean indicating if undo process ended up successfully,
        dict with all the executions). If undo process not executed, boolean set to None
    """
    rse = RegisterStepExecution(execution)
    execution.start_time = execution.start_time or get_now()
    cc = create_cmd_from_orchestration(orchestration, var_context, hosts=hosts, register=rse, executor=executor)

    # convert UUID into str as in_ filter does not handle UUID type
    all = [str(s) for s in hosts['all']]
    servers = Server.query.filter(Server.id.in_(all)).all()
    try:
        applicant = lock.lock(Scope.ORCHESTRATION, servers, execution.id)
    except errors.LockError as e:
        execution.end_time = get_now()
        execution.success = False
        execution.message = str(e)
        db.session.commit()
        raise
    try:
        execution.success = cc.invoke()
        if not execution.success and orchestration.undo_on_error:
            execution.undo_success = cc.undo()
        execution.end_time = get_now()
        db.session.commit()
    except Exception as e:
        current_app.logger.exception("Exception while executing invocation command")

    finally:
        lock.unlock(Scope.ORCHESTRATION, applicant=applicant, servers=servers)
    db.session.refresh(execution)
    return execution



import json

from prompt_toolkit.validation import Validator, ValidationError


class JSONValidator(Validator):

    def transform(self, text):
        return json.loads(text)

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.transform(text)
            except Exception as e:
                raise ValidationError(message=str(e))


class BoolValidator(Validator):
    true = ['true', 'yes', 'y']
    false = ['false', 'no', 'n']

    @classmethod
    def transform(cls, x: str):
        if x.strip().lower() in cls.true:
            return True
        elif x.strip().lower() in cls.false:
            return False
        else:
            raise ValueError

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.transform(text)
            except:
                raise ValidationError(message=f'Not a valid bool value ({",".join(self.true)}|{",".join(self.false)})')


class IntValidator(Validator):

    @staticmethod
    def transform(x: str):
        return int(x) if x is not None else x

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.transform(text)
            except:
                raise ValidationError(message='Not a valid integer')


class ChoiceValidator(Validator):

    def __init__(self, choices) -> None:
        super().__init__()
        self.choices = choices

    def validate(self, document):
        text = document.text

        if text and text not in self.choices:
            raise ValidationError(message='Invalid choice. Choose from ' + ', '.join(self.choices),
                                  cursor_position=0)


import ast


class ListValidator(Validator):

    @staticmethod
    def transform(x: str):
        v = ast.literal_eval(x) if x is not None else '[]'
        if v and not isinstance(v, list):
            raise ValueError()
        return v

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.transform(text)
            except Exception as e:
                raise ValidationError(message='Not a valid list. ' + str(e))



# noinspection PyUnresolvedReferences
from typing import Iterable

from flask import current_app


class ContextApp:
    __slots__ = ['_constructor', '_args', '_kwargs', '_app_container', 'app']

    def __init__(self, constructor, args=None, kwargs=None, app=None):
        self._constructor = constructor
        self._args = args or ()
        self._kwargs = kwargs or {}
        self._app_container = {}
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self._app_container.update({app: self._constructor(*self._args, **self._kwargs)})
        app.extensions[self._constructor.__name__.lower()] = self._app_container[app]

    def get_app(self, reference_app=None):
        """Helper method that implements the logic to look up an
        application."""

        if reference_app is not None:
            return reference_app

        if current_app:
            return current_app._get_current_object()

        if self.app:
            return self.app

        raise RuntimeError(
            'No application found. Either work inside a view function or push'
            ' an application context.'
        )

    @property
    def current(self):
        app = self.get_app()
        return self._app_container[app]


import os

from flask import request, g
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Software, Server, SoftwareServerAssociation
from dm.utils.helpers import md5
from dm.web import db, errors
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog
from dm.web.helpers import filter_query, check_param_in_uri
from dm.web.json_schemas import software_post, software_patch


def set_software_server(soft, server, path, recalculate_data=False):
    file = os.path.join(path, soft.filename)
    if not os.path.exists(file):
        raise errors.FileNotFound(file)

    if soft.size != os.path.getsize(file):
        return errors.GenericError(f"file is not of specified size", file=file, size=soft.size)
    if soft.checksum != md5(file):
        return errors.GenericError(f"checksum error on file", file=file)

    return SoftwareServerAssociation(software=soft, server=server, path=path)


# /software

class SoftwareList(Resource):

    @securizer
    @jwt_required
    @forward_or_dispatch
    def get(self):
        query = filter_query(Software, request.args)
        return [soft.to_json(servers=check_param_in_uri('servers')) for soft in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(software_post)
    @lock_catalog
    def post(self):
        json = request.get_json()

        file = json['file']
        if not os.path.exists(file):
            raise errors.FileNotFound(file)

        soft = Software(name=json['name'], version=json['version'], filename=os.path.basename(json['file']),
                        size=os.path.getsize(file), checksum=md5(file),
                        family=json.get('family', None))
        ssa = set_software_server(soft, g.server, os.path.dirname(json['file']))

        if not isinstance(ssa, SoftwareServerAssociation):
            return ssa
        db.session.add_all([soft, ssa])
        db.session.commit()
        return {'id': str(soft.id)}, 201


# /software/<software_id>
class SoftwareResource(Resource):
    @securizer
    @jwt_required
    @forward_or_dispatch
    def get(self, software_id):
        return Software.query.get_or_404(software_id).to_json()


# software/<software_id>/servers
class SoftwareServersResource(Resource):
    @securizer
    @jwt_required
    @forward_or_dispatch
    def get(self, software_id):
        soft = Software.query.get_or_404(software_id)
        return [ssa.server.to_json() for ssa in soft.ssas]

    # @forward_or_dispatch
    # @jwt_required
    # @securizer
    # @validate_schema(software_servers_put)
    # @lock_catalog
    # def put(self, software_id):
    #     json = request.get_json()
    #
    #     soft = Software.query.get_or_404(software_id)
    #
    #     # delete all associations
    #     soft.ssas = []
    #
    #     for ssa_json in json:
    #         server = Server.query.get_or_404(ssa_json['server_id'])
    #         ssa = SoftwareServerAssociation(software=soft, server=server, path=ssa_json['path'])
    #         db.session.add(ssa)
    #
    #     db.session.commit()
    #     return

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(software_patch)
    @lock_catalog
    def patch(self, software_id):
        json = request.get_json()

        soft = Software.query.get_or_404(software_id)
        server = Server.query.get_or_404(json['server_id'])

        ssa = set_software_server(soft, server, json['path'], recalculate_data=json.get('recalculate_data', False))
        db.session.add(ssa)
        db.session.commit()
        return {}, 204



# business functions related with business logic
import base64
import datetime
import json
import os
import subprocess
import typing as t

from flask import current_app
from flask_jwt_extended import create_access_token, get_jwt_identity
from sqlalchemy.orm.attributes import flag_modified

from dm import defaults
from dm.domain.entities import Server, bypass_datamark_update, Scope, Catalog
from dm.use_cases.lock import lock_scope
from dm.utils.helpers import get_distributed_entities
from dm.web import db, errors
from dm.web.network import get


def run_elevator(file, new_version, logger):
    logger.info(f"Upgrading to version {new_version}")
    stdout = open('elevator.out', 'a')
    cmd = ['python', 'elevator.py', 'upgrade', file, str(new_version)]
    logger.debug(f"Running command {' '.join(cmd)}")
    subprocess.Popen(cmd, stdin=None, stdout=stdout, stderr=stdout, close_fds=True, env=os.environ)
    stdout.close()


def get_software(server: Server, auth) -> t.Tuple[str, str]:
    resp, code = get(server, 'api_1_0.software_dimensigon', auth=auth)
    if code == 200:
        content = base64.b64decode(resp.get('content').encode('ascii'))

        file = os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon', resp.get('filename'))
        with open(file, 'wb') as fh:
            fh.write(content)
        return file, resp.get('version')
    else:
        return None, None


def upgrade_catalog(catalog, check_mismatch=True):
    de = get_distributed_entities()

    if check_mismatch:
        inside = set([name for name, cls in de])
        current_app.logger.debug(f'Actual entities: {inside}')

        outside = set(catalog.keys())
        current_app.logger.debug(f'Remote entities: {outside}')

        if len(inside ^ outside) > 0:
            raise errors.CatalogMismatch(inside, outside)

    with bypass_datamark_update():
        for name, cls in de:
            if name in catalog:
                if len(catalog[name]) > 0:
                    current_app.logger.debug(
                        f"Adding/Modifying new '{name}' entities: \n{json.dumps(catalog[name], indent=2, sort_keys=True)}")
                for dto in catalog[name]:
                    o = cls.from_json(dict(dto))
                    # force modification to update catalog last_modified_at
                    flag_modified(o, 'last_modified_at')
                    db.session.add(o)

        db.session.commit()


def upgrade_catalog_from_server(server):
    with lock_scope(Scope.UPGRADE, [server, Server.get_current()]):
        catalog_ver = db.session.query(db.func.max(Catalog.last_modified_at)).scalar()
        if catalog_ver:
            resp = get(server, 'api_1_0.catalog',
                       view_data=dict(data_mark=catalog_ver.strftime(defaults.DATEMARK_FORMAT)),
                       headers={'Authorization': 'Bearer ' + create_access_token(get_jwt_identity(),
                                                                                 expires_delta=datetime.timedelta(
                                                                                     seconds=15))})

            if resp.code and 199 < resp.code < 300:
                delta_catalog = resp.msg
                upgrade_catalog(delta_catalog)
            else:
                current_app.logger.error(f"Unable to get a valid response from server {server}: {resp}")



view_path_map = {'api_1_0.actiontemplatelist': '/api/v1.0/action_templates',
                'api_1_0.actiontemplateresource': '/api/v1.0/action_templates/<action_template_id>',
                'api_1_0.catalog': '/api/v1.0/catalog/<string:data_mark>',
                'api_1_0.catalog_update': '/api/v1.0/catalog',
                'api_1_0.events': '/api/v1.0/events/<event_id>',
                'api_1_0.granulelist': '/api/v1.0/granules',
                'api_1_0.home': '/api/v1.0/',
                'api_1_0.join': '/api/v1.0/join',
                'api_1_0.join_public': '/api/v1.0/join/public',
                'api_1_0.launch_command': '/api/v1.0/launch/command',
                'api_1_0.launch_operation': '/api/v1.0/launch/operation',
                'api_1_0.launch_orchestration': '/api/v1.0/launch/orchestration/<orchestration_id>',
                'api_1_0.locker': '/api/v1.0/locker',
                'api_1_0.locker_lock': '/api/v1.0/locker/lock',
                'api_1_0.locker_prevent': '/api/v1.0/locker/prevent',
                'api_1_0.locker_unlock': '/api/v1.0/locker/unlock',
                'api_1_0.loglist': '/api/v1.0/log',
                'api_1_0.logresource': '/api/v1.0/log/<log_id>',
                'api_1_0.orchestrationexecutionrelationship': '/api/v1.0/orchestrations/<orchestration_id>/executions',
                'api_1_0.orchestrationlist': '/api/v1.0/orchestrations',
                'api_1_0.orchestrationresource': '/api/v1.0/orchestrations/<orchestration_id>',
                'api_1_0.orchestrationfull': '/api/v1.0/orchestrations/full',
                'api_1_0.orchexecstepexecrelationship': '/api/v1.0/orchestration_executions/<execution_id>/step_executions',
                'api_1_0.orchexecutionlist': '/api/v1.0/orchestration_executions',
                'api_1_0.orchexecutionresource': '/api/v1.0/orchestration_executions/<execution_id>',
                'api_1_0.routes': '/api/v1.0/routes',
                'api_1_0.send': '/api/v1.0/send',
                'api_1_0.serverlist': '/api/v1.0/servers',
                'api_1_0.serverresource': '/api/v1.0/servers/<server_id>',
                'api_1_0.software_dimensigon': '/api/v1.0/software/dimensigon',
                'api_1_0.softwarelist': '/api/v1.0/software',
                'api_1_0.softwareresource': '/api/v1.0/software/<software_id>',
                'api_1_0.softwareserversresource': '/api/v1.0/software/<software_id>/servers',
                'api_1_0.stepexecutionlist': '/api/v1.0/step_executions',
                'api_1_0.stepexecutionresource': '/api/v1.0/step_executions/<execution_id>',
                'api_1_0.steplist': '/api/v1.0/steps',
                'api_1_0.steprelationshipchildren': '/api/v1.0/steps/<step_id>/relationship/children',
                'api_1_0.steprelationshipparents': '/api/v1.0/steps/<step_id>/relationship/parents',
                'api_1_0.stepresource': '/api/v1.0/steps/<step_id>',
                'api_1_0.transferlist': '/api/v1.0/transfers',
                'api_1_0.transferresource': '/api/v1.0/transfers/<transfer_id>',
                'api_1_0.userlist': '/api/v1.0/users',
                'api_1_0.userresource': '/api/v1.0/users/<user_id>',
                'root.fresh_login': '/fresh-login',
                'root.healthcheck': '/healthcheck',
                'root.home': '/',
                'root.login': '/login',
                'root.ping': '/ping',
                'root.refresh': '/refresh'
                }



import inspect

from prompt_toolkit import prompt


def resolve_callables(kwargs, params):
    prompt_kwargs = dict(kwargs)
    for kk, vv in kwargs.items():
        if callable(vv):
            sig = inspect.signature(vv)
            callable_kwargs = {kw: params.get(kw, None) for kw in sig.parameters.keys()}
            prompt_kwargs[kk] = vv(**callable_kwargs)
    return prompt_kwargs


def prompt_parameter(parameter, entity, form, parent_prompt) -> bool:
    prompt_kwargs = resolve_callables(form[parameter], entity)

    text = prompt(
        f"{parent_prompt}.{parameter}>{'>' if prompt_kwargs.get('multiline', False) else ''} ",
        default=str(entity.get(parameter, '')) if entity.get(parameter, '') != '' else '',
        **prompt_kwargs)
    if text == '':
        text = None
    elif text == "''" or text == '""':
        text = ""
    else:
        if 'validator' in prompt_kwargs and hasattr(prompt_kwargs['validator'], 'transform'):
            text = prompt_kwargs['validator'].transform(text)
    if text != entity.get(parameter, None):
        entity[parameter] = text
        return True
    else:
        return False



import atexit
import datetime
import os
import threading
import typing as t

from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask, g, _app_ctx_stack, _request_ctx_stack
from flask_jwt_extended import JWTManager
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData

from config import config_by_name
from dm.utils.event_handler import EventHandler
from dm.web import errors
from .extensions.flask_executor.executor import Executor
from .helpers import BaseQueryJSON, run_in_background
from ..utils.helpers import get_now


def scopefunc():
    try:
        return str(id(_app_ctx_stack.top.app)) + str(threading.get_ident()) + str(id(_request_ctx_stack.top.request))
    except:
        try:
            return str(id(_app_ctx_stack.top.app)) + str(threading.get_ident())
        except:
            return str(threading.get_ident())


meta = MetaData(naming_convention={
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
})

db = SQLAlchemy(query_class=BaseQueryJSON, metadata=meta, session_options=dict(scopefunc=scopefunc))
# db = SQLAlchemy(query_class=BaseQueryJSON)
jwt = JWTManager()
executor = Executor()


class DimensigonApp(Flask):

    def start_background_tasks(self):
        from ..use_cases.log_sender import LogSender
        from dm.web.background_tasks import process_catalog_route_table, process_get_new_version_from_gogs
        if not self.config['TESTING'] or os.getenv('WERKZEUG_RUN_MAIN') == 'true':
            if self.extensions.get('scheduler') is None and self.config['SCHEDULER']:
                bs = BackgroundScheduler()
                self.extensions['scheduler'] = bs
                ls = LogSender()
                self.extensions['log_sender'] = ls
                bs.start()

                if self.config.get('AUTOUPGRADE'):
                    bs.add_job(func=process_get_new_version_from_gogs, args=(self,), trigger="interval", hours=6)
                bs.add_job(func=process_catalog_route_table, name="catalog & route upgrade", args=(self,),
                           trigger="interval", minutes=5,
                           next_run_time=get_now() + datetime.timedelta(seconds=5))
                bs.add_job(func=run_in_background, name="log_sender", args=(ls.send_new_data, self),
                           trigger="interval",
                           minutes=2, next_run_time=get_now() + datetime.timedelta(seconds=30))

                # Shut down the scheduler when exiting the app
                atexit.register(lambda: bs.shutdown())

    def run(self, host=None, port=None, debug=None, load_dotenv=True, **options):
        from dm.domain.entities.bootstrap import set_initial
        set_initial(server=False, user=False)
        db.session.commit()
        self.start_background_tasks()
        super(DimensigonApp, self).run(host=host, port=port, debug=debug, load_dotenv=load_dotenv, **options)


def _initialize_blueprint(app):
    from dm.web.routes import root_bp
    from dm.web.api_1_0 import api_bp as api_1_0_bp

    app.register_blueprint(root_bp)
    handle_exception = app.handle_exception
    handle_user_exception = app.handle_user_exception
    app.register_blueprint(api_1_0_bp)
    app.handle_exception = handle_exception
    app.handle_user_exception = handle_user_exception


def _initialize_errorhandlers(application):
    '''
    Initialize error handlers
    '''
    from dm.web.errors import bp_errors
    application.register_blueprint(bp_errors)


def create_app(config_name):
    app = DimensigonApp('dm')
    if isinstance(config_name, t.Mapping):
        app.config.from_mapping(config_name)
    elif config_name in config_by_name:
        app.config.from_object(config_by_name[config_name])
        config_by_name[config_name].init_app(app)
    else:
        app.config.from_object(config_name)
        if hasattr(config_name, 'init_app'):
            config_name.init_app(app)

    # EXTENSIONS
    db.init_app(app)
    jwt.init_app(app)
    executor.init_app(app)
    app.events = EventHandler()

    app.before_first_request_funcs = [app.start_background_tasks]
    app.before_request(load_global_data_into_context)
    _initialize_blueprint(app)
    _initialize_errorhandlers(app)

    return app


@jwt.user_loader_callback_loader
def user_loader_callback(identity):
    from ..domain.entities import User
    return User.query.get(identity)


def load_global_data_into_context():
    from dm.domain.entities import Server, Dimension
    g.server = Server.get_current()
    g.dimension = Dimension.get_current()



import base64
import logging
import os
import typing as t

from dm.domain.entities import Log, Server
from dm.domain.entities.log import Mode
from dm.use_cases.helpers import get_auth_root
from dm.utils import asyncio
from dm.utils.pygtail import Pygtail
from dm.utils.typos import Id
from dm.web.network import async_post

MAX_LINES = 10000

logger = logging.getLogger('dm.background.log_sender')


class _PygtailBuffer(Pygtail):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._buffer = None

    def fetch(self):
        if self._buffer:
            return self._buffer
        else:
            self._buffer = ''.join(self.readlines(max_lines=MAX_LINES))
        return self._buffer

    def update_offset_file(self):
        self._buffer = None
        super().update_offset_file()


class LogSender:

    def __init__(self):
        self.buffer_data = None
        self.sending_data = True
        self._mapper: t.Dict[Id, t.List[_PygtailBuffer]] = {}

    @property
    def logs(self):
        return Log.query.filter_by(source_server=Server.get_current()).all()

    def update_mapper(self):
        logs = self.logs
        id2log = {log.id: log for log in logs}
        # remove logs
        for log_id in list(self._mapper.keys()):
            if log_id not in id2log:
                del self._mapper[log_id]

        # add new logs
        for log in logs:
            if log.id not in self._mapper:
                self._mapper[log.id] = []
            self.update_pytail_objects(log, self._mapper[log.id])

    def update_pytail_objects(self, log: Log, pytail_list: t.List):
        if os.path.isfile(log.target):
            if len(pytail_list) == 0:
                filename = '.' + os.path.basename(log.target) + '.offset'
                path = os.path.dirname(log.target)
                offset_file = os.path.join(path, filename)
                pytail_list.append(
                    _PygtailBuffer(file=log.target, offset_mode='manual', offset_file=offset_file))
        else:
            for folder, dirnames, filenames in os.walk(log.target):
                for filename in filenames:
                    if log._re_include.search(filename) and not log._re_exclude.search(filename):
                        file = os.path.join(folder, filename)
                        offset_file = os.path.join(folder, '.' + filename + '.offset')
                        if not any(map(lambda p: p.file == file, pytail_list)):
                            pytail_list.append(
                                _PygtailBuffer(file=file, offset_mode='manual', offset_file=offset_file))
                if not log.recursive:
                    break
                new_dirnames = []
                for dirname in dirnames:
                    if log._re_include.search(dirname) and not log._re_exclude.search(dirname):
                        new_dirnames.append(dirname)
                dirnames[:] = new_dirnames

    async def send_new_data(self):
        self.update_mapper()
        tasks = []
        auth = get_auth_root()
        for log_id, pb in self._mapper.items():
            log = Log.query.get(log_id)
            for pytail in pb:
                data = pytail.fetch()
                data = data.encode() if isinstance(data, str) else data
                if log.mode == Mode.MIRROR:
                    file = pytail.file
                elif log.mode == Mode.REPO_ROOT:
                    path_to_remove = os.path.dirname(log.target)
                    relative = os.path.relpath(pytail.file, path_to_remove)
                    file = os.path.join('{LOG_REPO}', relative)
                elif log.mode == Mode.FOLDER:
                    path_to_remove = os.path.dirname(log.target)
                    relative = os.path.relpath(pytail.file, path_to_remove)
                    file = os.path.join(log.dest_folder, relative)
                else:
                    def get_root(dirname):
                        new_dirname = os.path.dirname(dirname)
                        if new_dirname == dirname:
                            return dirname
                        else:
                            return get_root(new_dirname)

                    relative = os.path.relpath(pytail.file, get_root(pytail.file))
                    file = os.path.join('{LOG_REPO}', relative)
                task = asyncio.create_task(
                    async_post(log.destination_server, 'api_1_0.logresource', view_data={'log_id': str(log_id)},
                               json={"file": file, 'data': base64.b64encode(data).decode('ascii')},
                               auth=auth))

                tasks.append((task, pytail, log))
                logger.debug(f"Sending data from '{pytail.file}' to '{log.destination_server}'")

        for task, pytail, log in tasks:
            response = await task
            if response.ok:
                pytail.update_offset_file()
                logger.debug(f"Updated offset from '{pytail.file}'")
            else:
                if response.exception:
                    logger.exception(
                        f"Unable to send log information from '{pytail.file}' to '{log.destination_server}'",
                        exc_info=response)
                else:
                    logger.error(
                        f"Unable to send log information from '{pytail.file}' to '{log.destination_server}'. Error:"
                        f"{response[1]}, {response[0]}")



import base64
import concurrent
import copy
import datetime as dt
import functools
import ipaddress
import math
import os
import pickle
import random
import re
import time
import traceback
import typing as t
import uuid
from collections import OrderedDict

from flask import request, current_app, jsonify, g
from flask_jwt_extended import jwt_required, get_jwt_identity, create_access_token, current_user
from pkg_resources import parse_version

import dm
from dm import defaults as d, defaults
from dm.domain.entities import Software, Server, SoftwareServerAssociation, Catalog, Route, StepExecution, \
    Orchestration, OrchExecution, User, ActionTemplate, ActionType
from dm.network.auth import HTTPBearerAuth
from dm.utils import asyncio, subprocess
from dm.utils.dag import DAG
from dm.utils.event_handler import Event
from dm.utils.helpers import get_distributed_entities, is_iterable_not_string, md5, get_now
from dm.utils.typos import Id
from dm.utils.var_context import VarContext
from dm.web import db, executor, errors
from dm.web.api_1_0 import api_bp
from dm.web.async_functions import deploy_orchestration, async_send_file
from dm.web.background_tasks import update_table_routing_cost, _get_neighbour_catalog_data_mark, upgrade_version, \
    update_catalog
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog
from dm.web.helpers import check_param_in_uri, normalize_hosts, search
from dm.web.json_schemas import launch_command_post, routes_post, routes_patch, \
    launch_orchestration_post, send_post, orchestration_full
from dm.web.network import async_post
from dm.web.network import post, get

if t.TYPE_CHECKING:
    from dm.use_cases.operations import IOperationEncapsulation


@api_bp.route('/')
def home():
    return "API v1.0 documentation page"


@api_bp.route('/join/public', methods=['GET'])
@jwt_required
def join_public():
    if current_user.user == User.get_by_user('join').user:
        return g.dimension.public.save_pkcs1(), 200, {'content-type': 'application/octet-stream'}
    else:
        return {}, 401


@api_bp.route('/join', methods=['POST'])
@securizer
@jwt_required
@lock_catalog
def join():
    if current_user.user == User.get_by_user('join').user:
        js = request.get_json()
        current_app.logger.debug(f"New server wanting to join: {js}")
        s = Server.from_json(js)
        external_ip = ipaddress.ip_address(request.remote_addr)
        if not external_ip.is_loopback and external_ip not in [g.ip for g in s.gates]:
            for port in set([g.port for g in s.gates]):
                s.add_new_gate(external_ip, port, hidden=True)

        Route(destination=s, cost=0)
        db.session.add(s)
        db.session.commit()
        catalog = fetch_catalog(defaults.INITIAL_DATEMARK)
        catalog.update(Dimension=g.dimension.to_json())
        catalog.update(me=str(Server.get_current().id))
        return catalog, 200
    else:
        return '', 401


@api_bp.route('/send', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(send_post)
def send():
    def search_cost(ssa, route_list):
        cost = [route['cost'] for route in route_list if str(ssa.server.id) == route['destination_id']]
        if cost:
            if cost[0] is None:
                cost = 999999
            else:
                cost = cost[0]
        else:
            cost = 999999
        return cost

    auth = HTTPBearerAuth(create_access_token(get_jwt_identity(), expires_delta=dt.timedelta(
        minutes=30)))
    # Validate Data
    json_data = request.get_json()

    dest_server = Server.query.get_or_404(json_data['dest_server_id'])

    if 'software_id' in json_data:
        software = Software.query.get_or_404(json_data['software_id'])

        ssa = SoftwareServerAssociation.query.filter_by(server=g.server, software=software).one_or_none()
        # if current server does not have the software, forward request to the closest server who has it
        if not ssa:
            resp = get(dest_server, 'api_1_0.routes', auth=auth, timeout=5)
            if resp.code == 200:
                ssas = copy.copy(software.ssas)
                ssas.sort(key=functools.partial(search_cost, route_list=resp.msg['route_list']))
            # unable to get route cost, we take the first option we have
            else:
                ssas = random.shuffle(list(software.ssas))
            if len(ssas) == 0:
                raise errors.NoSoftwareServer(software_id=str(software.id))
            server = ssas[0].server  # closest server from dest_server who has the software

            resp = post(server, 'api_1_0.send', json=json_data, auth=auth)
            resp.raise_if_not_ok()
            return resp.msg, resp.code
        else:

            file = os.path.join(ssa.path, software.filename)
            if not os.path.exists(file):
                raise errors.FileNotFound(file)
            size = ssa.software.size
    else:
        file = json_data['file']
        if os.path.exists(file):
            size = os.path.getsize(file)
            checksum = md5(json_data.get('file'))
        else:
            raise errors.FileNotFound(file)

    chunk_size = min(json_data.get('chunk_size', d.CHUNK_SIZE), d.CHUNK_SIZE) * 1024
    max_senders = min(json_data.get('max_senders', d.MAX_SENDERS), d.MAX_SENDERS)
    chunks = math.ceil(size / chunk_size)

    if 'software_id' in json_data:
        json_msg = dict(software_id=str(software.id), num_chunks=chunks,
                        dest_path=json_data.get('dest_path', current_app.config['SOFTWARE_REPO']))
    else:
        json_msg = dict(filename=os.path.basename(json_data.get('file')), size=size, checksum=checksum,
                        num_chunks=chunks, dest_path=json_data.get('dest_path', current_app.config['SOFTWARE_REPO']))
    if 'force' in json_data:
        json_msg['force'] = json_data['force']

    resp = post(dest_server, 'api_1_0.transferlist', json=json_msg, auth=auth)
    resp.raise_if_not_ok()

    transfer_id = resp.msg.get('id')
    current_app.logger.debug(
        f"Transfer {transfer_id} created. Sending {file} to {dest_server}:{json_data.get('dest_path')}.")

    if json_data.get('background', True):
        executor.submit(asyncio.run,
                        async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                        chunk_size=chunk_size, max_senders=max_senders, auth=auth))
    else:
        asyncio.run(async_send_file(dest_server=dest_server, transfer_id=transfer_id, file=file,
                                    chunk_size=chunk_size, max_senders=max_senders, auth=auth))

    if json_data.get('include_transfer_data', False):
        resp = get(dest_server, "api_1_0.transferresource", view_data=dict(transfer_id=transfer_id), auth=auth)
        if resp.code == 200:
            msg = resp.msg
        else:
            resp.raise_if_not_ok()
    else:
        msg = {'transfer_id': transfer_id}
    return msg, 202 if json_data.get('background', True) else 201


@api_bp.route('/software/dimensigon', methods=['GET'])
@forward_or_dispatch
@jwt_required
@securizer
def software_dimensigon():
    # sends the last software
    repo = current_app.config['SOFTWARE_REPO']
    max_version = None
    max_file = None
    for file in os.listdir(os.path.join(repo, 'dimensigon')):
        if 'dimensigon-' in file:
            m = re.search(r'v?\d+\.\d+[.-][ab]?\d+', file)
            if m and parse_version(m.group()) >= parse_version(max_version or dm.__version__):
                max_version = m.group()
                max_file = file
    if max_file:
        with open(os.path.join(repo, 'dimensigon', max_file), 'rb') as fh:
            return {'filename': max_file, 'version': max_version,
                    'content': base64.b64encode(fh.read()).decode('ascii')}, 200
    else:
        return {}, 204


@api_bp.route('/catalog', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
def catalog_update():
    data = asyncio.run(_get_neighbour_catalog_data_mark())
    # check version upgrade before catalog upgrade to match database revision
    if not upgrade_version(data):
        update_catalog(data)
    db.session.commit()
    return {}, 204


@api_bp.route('/catalog/<string:data_mark>', methods=['GET', 'POST'])
@forward_or_dispatch
@jwt_required
@securizer
def catalog(data_mark):
    # Input Validation
    if data_mark == 'initial':
        data_validated = dt.datetime(dt.MINYEAR, 1, 1)
    else:
        try:
            data_validated = dt.datetime.strptime(data_mark, defaults.DATEMARK_FORMAT)
        except Exception as e:
            return {'error': f'Invalid Data Mark: {e}'}, 400

    return fetch_catalog(data_validated)


def fetch_catalog(data_mark):
    data = {}
    for name, obj in get_distributed_entities():
        c = Catalog.query.get(name)
        repo_data = obj.query.filter(obj.last_modified_at > data_mark).all()
        if name == 'User':
            data.update({name: [e.to_json(password=True) for e in repo_data]})
        else:
            data.update({name: [e.to_json() for e in repo_data]})
    return data


@api_bp.route('/routes', methods=['GET', 'POST', 'PATCH'])
@jwt_required
@securizer
@validate_schema(POST=routes_post, PATCH=routes_patch)
def routes():
    if request.method == 'GET':

        route_table = []
        for route in Route.query.join(Server.route).order_by(
                Server.name).all():
            route_table.append(route.to_json(human=check_param_in_uri('human')))
        data = {'route_list': route_table}
        if check_param_in_uri('human'):
            data.update(server=str(g.server))
        else:
            data.update(server_id=str(g.server.id))
        return data

    elif request.method == 'POST':
        data = request.get_json()
        new_routes = update_table_routing_cost(**data)
        db.session.commit()

        # if len(new_routes) > 0:
        #     def func_patch(app, *args, **kwargs):
        #         with app.app_context():
        #             patch(*args, **kwargs)
        #
        #     msg = {'server_id': str(Server.get_current().id),
        #            'route_list': [
        #                dict(destination_id=str(d.id),
        #                     proxy_server_id=str(getattr(r.proxy_server, 'id')) if getattr(r.proxy_server, 'id',
        #                                                                                   None) else None,
        #                     gate_id=str(getattr(r.gate, 'id')) if getattr(r.gate, 'id', None) else None,
        #                     cost=r.cost)
        #                for d, r in new_routes.items()
        #            ]}
        #
        #     for s in Server.get_neighbours():
        #         th = threading.Thread(target=func_patch, args=(current_app._get_current_object(), s, 'api_1_0.routes'),
        #                               kwargs={'json': msg,
        #                                       'headers': dict(Authorization=request.headers['Authorization'])})
        #         th.daemon = True
        #         th.start()

    # elif request.method == 'PATCH':
    #     data = request.get_json()
    #     current_app.logger.debug(f"New routes recived: {json.dumps(data, indent=4)}")
    #     likely_proxy_server = Server.query.get(data.get('server_id'))
    #     new_routes = []
    #     if not likely_proxy_server:
    #         return {"error": f"Server id '{data.get('server_id')}' not found"}, 404
    #     for new_route in data.get('route_list', []):
    #         target_server = Server.query.get(new_route.get('destination_id'))
    #         if target_server is None:
    #             current_app.logger.debug(f"Destination server unknown {new_route.get('destination_id')}")
    #             continue
    #         if target_server == g.server:
    #             # check if server has detected me as a neighbour
    #             if new_route.get('cost') == 0:
    #                 # server may be created without route (backward compatibility)
    #                 if likely_proxy_server.route is None:
    #                     likely_proxy_server.route = Route(destination=likely_proxy_server)
    #                 # check if I do not have it as a neighbour yet
    #                 if likely_proxy_server.route.cost != 0:
    #                     for gate in likely_proxy_server.gates:
    #                         if check_host(gate.dns or str(gate.ip), gate.port, timeout=1, retry=3, delay=0.5):
    #                             likely_proxy_server.route.proxy_server = None
    #                             likely_proxy_server.route.gate = gate
    #                             likely_proxy_server.route.cost = 0
    #                             new_routes.append(likely_proxy_server.route)
    #                             break
    #         else:
    #             # server may be created without route (backward compatibility)
    #             if target_server.route is None:
    #                 target_server.route = Route(destination=target_server)
    #             # process routes whose proxy_server is not me
    #             if str(g.server.id) != new_route.get('proxy_server_id'):
    #                 if target_server.route.proxy_server == likely_proxy_server and new_route.get('cost') is None:
    #                     cost, time = None, None
    #                 else:
    #                     cost, time = ping_server(target_server, g.server)  # check my route
    #                 if cost is not None:
    #                     # if new route has less cost than actual route, take it as my new route
    #                     if ((new_route.get('cost') or 999999) + 1) < cost:
    #                         target_server.route.proxy_server = likely_proxy_server
    #                         target_server.route.gate = None
    #                         target_server.route.cost = new_route.get('cost') + 1
    #                         new_routes.append(target_server.route)
    #                 else:
    #                     # if new route reaches the destination take it as a new one
    #                     if new_route.get('cost') is not None:
    #                         target_server.route.proxy_server = likely_proxy_server
    #                         target_server.route.gate = None
    #                         target_server.route.cost = new_route.get('cost') + 1
    #                     else:
    #                         # neither my route and the new route has access to the destination
    #                         target_server.route.gate, target_server.route.proxy_server, target_server.route.cost = None, None, None
    #                     new_routes.append(target_server.route)
    #
    #     # # Seek my routes whose gateway is the likely_proxy_server
    #     # for target_server in Server.query.join(Route.destination).filter(
    #     #         Route.proxy_server == likely_proxy_server).all():
    #     #     for new_route in data.get('route_list', []):
    #     #         if new_route.get('cost'):
    #     #             target_server.route.cost = new_route.get('cost') + 1
    #     #         else:
    #     #             target_server.route.gate, target_server.route.proxy_server, target_server.route.cost = None, None, None
    #     #         new_routes.append(target_server.route)
    #     db.session.commit()
    #
    #     # send new information in background
    #     if new_routes:
    #         def func_patch(app, *args, **kwargs):
    #             with app.app_context():
    #                 patch(*args, **kwargs)
    #
    #         msg = {'server_id': str(g.server.id),
    #                'route_list': [
    #                    r.to_json()
    #                    for r in new_routes]}
    #
    #         for s in Server.get_neighbours():
    #             if s != likely_proxy_server:
    #                 th = threading.Thread(target=func_patch,
    #                                       args=(current_app._get_current_object(), s, 'api_1_0.routes'),
    #                                       kwargs={'json': msg,
    #                                               'headers': dict(Authorization=request.headers['Authorization'])})
    #                 th.daemon = True
    #                 th.start()

    return {}, 204


def run_command_and_callback(operation: 'IOperationEncapsulation', var_context, source: Id, execution: Id, jwt_identity,
                             timeout=None):
    cp = operation.execute(var_context=var_context, timeout=timeout)

    execution = StepExecution.query.get(execution)
    source = Server.query.get(source)
    execution.load_completed_result(cp)
    try:
        db.session.commit()
    except Exception as e:
        current_app.logger.exception(f"Error on commit for execution {execution.id}")
    data = dict(**execution.to_json(),
                                var_context=base64.b64encode(pickle.dumps(var_context)).decode('ascii'))
    resp, code = post(server=source, view_or_url='api_1_0.events', view_data={'event_id': str(execution.id)},
                      json=data,
                      auth=HTTPBearerAuth(
                          create_access_token(jwt_identity, expires_delta=dt.timedelta(seconds=30))))
    if code != 202:
        current_app.logger.error(f"Error while sending result for execution {execution.id}: {code}, {resp}")
    return data


@api_bp.route('/launch/operation', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
def launch_operation():
    data = request.get_json()
    operation = pickle.loads(base64.b64decode(data['operation'].encode('ascii')))
    var_context = pickle.loads(base64.b64decode(data['var_context'].encode('ascii')))
    orch_exec_json = data['execution']
    orch_exec = OrchExecution.from_json(orch_exec_json)

    e = StepExecution(server=g.server, step_id=data.get('step_id'), orch_execution=orch_exec, params=dict(var_context),
                      start_time=get_now())
    db.session.add_all([orch_exec, e])
    db.session.commit()
    future = executor.submit(run_command_and_callback, operation, var_context, g.source.id, e.id,
                             get_jwt_identity(),
                             timeout=data.get('timeout'))
    try:
        r = future.result(1)
    except concurrent.futures.TimeoutError:
        return {'execution_id': str(e.id)}, 202
    except Exception as e:
        current_app.logger.exception(
            f"Exception got when executing step {data.get('step_id')}. See logs for more information")
        return {'error': f"Exception got when executing step {data.get('step_id')}. See logs for more information"}, 500
    else:
        return r, 200





@api_bp.route('/launch/orchestration/<orchestration_id>', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(launch_orchestration_post)
def launch_orchestration(orchestration_id):
    orchestration = Orchestration.query.get_or_404(orchestration_id)
    params = request.get_json().get('params', {})
    hosts = request.get_json().get('hosts')

    a = set(orchestration.target)
    if not isinstance(hosts, dict):
        hosts = dict(all=hosts)
    b = set(hosts.keys())
    c = a - b
    if len(c) > 0:
        raise errors.TargetUnspecified(c)
    c = b - a
    if len(c) > 0:
        raise errors.TargetNotNeeded(c)

    not_found = normalize_hosts(hosts)
    if not_found:
        raise errors.ServerNormalizationError(not_found)

    for target, target_hosts in hosts.items():
        if len(target_hosts) == 0:
            raise errors.EmptyTarget(target)
    # check param entries
    # rest = orchestration.user_parameters - set(params.keys())
    # if rest:
    #     rest = list(rest)
    #     rest.sort()
    #     return {'error': f"Parameter(s) not specified: {', '.join(rest)}"}, 404

    if not orchestration.steps:
        return errors.GenericError('orchestration does not have steps to execute', orchestration_id=orchestration_id)

    execution_id = str(uuid.uuid4())

    vc = VarContext(globals=dict(execution_id=execution_id, executor_id=get_jwt_identity()), initials=params)

    if request.get_json().get('background', True):
        future = executor.submit(deploy_orchestration, orchestration=orchestration.id, var_context=vc, hosts=hosts)
        try:
            r = future.result(1)
        except concurrent.futures.TimeoutError:
            return {'execution_id': execution_id}, 202
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
        else:
            return jsonify(r), 200
    else:
        try:
            orch_exe = deploy_orchestration(orchestration=orchestration.id, var_context=vc, hosts=hosts)
        except Exception as e:
            current_app.logger.exception(f"Exception got when executing orchestration {orchestration}")
            raise
        else:
            return orch_exe.to_json(add_step_exec=True, human=check_param_in_uri('human')), 200

async def parallel_command_run(servers, data, auth):
    server_responses = {}
    for server in servers:
        server_responses[str(server.id)] = asyncio.create_task(
            async_post(server, 'api_1_0.launch_command', json=data, auth=auth, timeout=data.get('timeout', None)))

    for server in servers:
        server_responses[str(server.id)] = await server_responses[str(server.id)]
    return server_responses


def wrap_sudo(user, cmd):
    if isinstance(user, User):
        username = user.name
    else:
        username = user
    if isinstance(cmd, str):
        return f"sudo -Siu {username} {cmd}"
    else:
        args = ["sudo", "-Siu", username]
        args.extend(cmd)
        return args


@api_bp.route('/launch/command', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(launch_command_post)
def launch_command():
    data = request.get_json()

    server_list = []
    if 'hosts' in data:
        not_found = []
        servers = Server.query.all()
        if data['hosts'] == 'all':
            server_list = servers
        elif is_iterable_not_string(data['hosts']):
            for vv in data['hosts']:
                sl = search(vv, servers)
                if len(sl) == 0:
                    not_found.append(vv)
                else:
                    server_list.extend(sl)
        else:
            sl = search(data['hosts'], servers)
            if len(sl) == 0:
                not_found.append(data['hosts'])
            else:
                server_list.extend(sl)
        if not_found:
            return {'error': "Following granules or ids did not match to any server: " + ', '.join(not_found)}, 404
    else:
        server_list.append(g.server)

    if re.search(r'rm\s+((-\w+|--[-=\w]*)\s+)*(-\w*[rR]\w*|--recursive)', data['command']):
        return {'error': 'rm with recursion is not allowed'}, 403
    data.pop('hosts', None)
    start = None
    resp = {}
    username = getattr(current_user, 'user', None)
    if not username:
        raise errors.EntityNotFound('User', current_user)
    cmd = wrap_sudo(username, data['command'])
    if g.server in server_list:
        start = time.time()
        server_list.pop(server_list.index(g.server))
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE if data.get('input', None) else None,
                                stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                                shell=True, close_fds=True)

    resp_data = {}
    if server_list:
        resp = asyncio.run(
            parallel_command_run(server_list, data, auth=HTTPBearerAuth(create_access_token(get_jwt_identity()))))
        for s, r in resp.items():
            if r[1] == 200:
                if s in r[0]:
                    resp_data[s] = r[0][s]
                else:
                    resp_data[s] = r[0]
            else:
                if r[1]:
                    resp_data[s] = {
                        'error': {'status_code': r[1], 'response': r[0] if isinstance(r[0], str) else str(r[0])}}
                else:
                    if current_app.config['DEBUG']:
                        f_tb = traceback.format_exception(type(r[0]), r[0], r[0].__traceback__)
                    else:
                        f_tb = traceback.format_exception_only(type(r[0]), r[0], r[0])
                    resp_data[s] = {'error': ''.join(f_tb)}

    if start:
        timeout = data.get('timeout', defaults.TIMEOUT_COMMAND)
        try:
            outs, errs = proc.communicate(input=(data.get('input', '') or '').encode(), timeout=timeout - (time.time() - start))
        except (TimeoutError, subprocess.TimeoutExpired):
            proc.kill()
            try:
                outs, errs = proc.communicate(timeout=1)
            except:
                resp_data[str(g.server.id)] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds. Unable to communicate with the process."}
            else:
                resp_data[str(g.server.id)] = {
                    'error': f"Command '{cmd}' timed out after {timeout} seconds",
                    'stdout': outs.split('\n'), 'stderr': errs.split('\n')}
        except Exception as e:
            current_app.logger.exception("Exception raised while trying to run command")
            resp_data[str(g.server.id)] = {
                'error': traceback.format_exc() if current_app.config['DEBUG'] else traceback.format_exception_only()}
        else:
            resp_data[str(g.server.id)] = {'stdout': outs.decode().split('\n'), 'stderr': errs.decode().split('\n'),
                                           'returncode': proc.returncode}
    if check_param_in_uri("human"):
        convert_human = {}
        for s_id, r in resp_data.items():
            convert_human[Server.query.get(s_id).name] = r
        resp_data = convert_human
    resp_data['cmd'] = cmd
    resp_data['input'] = data.get('input', None)
    return resp_data, 200


@api_bp.route('/events/<event_id>', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
def events(event_id):
    e = Event(event_id, data=request.get_json())
    current_app.events.dispatch(e)
    return {}, 202


@api_bp.route('/orchestrations/full', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(orchestration_full)
@lock_catalog
def orchestrations_full():
    json_data = request.get_json()
    json_steps = json_data.pop('steps')
    generated_version = False
    if 'version' not in json_data:
        generated_version = True
        json_data['version'] = Orchestration.query.filter_by(name=json_data['name']).count() + 1
    o = Orchestration(**json_data)
    db.session.add(o)
    resp_data = {'id': str(o.id)}
    if generated_version:
        resp_data.update(version=o.version)

    # reorder steps in order of dependency
    id2step = {str(s['id']): s for s in json_steps}

    dag = DAG()
    for s in json_steps:
        step_id = str(s['id'])
        if s['undo'] and len(s.get('parent_step_ids', [])) == 0:
            raise errors.UndoStepWithoutParent(step_id)
        dag.add_node(step_id)
        for p_s_id in s.get('parent_step_ids', []):
            dag.add_edge(str(p_s_id), step_id)

    if dag.is_cyclic():
        raise errors.CycleError

    new_steps = []
    for step_id in dag.ordered_nodes:
        step = id2step[step_id]
        new_steps.append(step)
    # end reorder steps in order of dependency

    rid2step = OrderedDict()
    dependencies = {}
    for json_step in new_steps:
        rid = str(json_step.pop('id', None))
        if rid is not None and rid in rid2step.keys():
            raise errors.DuplicatedId(rid)
        if 'action_template_id' in json_step:
            json_step['action_template'] = ActionTemplate.query.get_or_404(json_step.pop('action_template_id'))
        elif 'action_type' in json_step:
            json_step['action_type'] = ActionType[json_step.pop('action_type')]
        dependencies[rid] = {'parent_step_ids': [str(p_id) for p_id in json_step.pop('parent_step_ids', [])]}
        s = o.add_step(**json_step)
        db.session.add(s)
        if rid:
            rid2step[rid] = s

        continue

    # process dependencies
    for rid, dep in dependencies.items():
        step = rid2step[rid]
        parents = []
        for p_s_id in dep['parent_step_ids']:
            if p_s_id in rid2step:
                parents.append(rid2step[p_s_id])
        o.set_parents(step, parents)

    db.session.commit()

    # send new ids in order of appearance at beginning
    new_id_steps = []
    for rid in rid2step.keys():
        new_id_steps.append(str(rid2step[rid].id))
    resp_data.update({'step_ids': new_id_steps})
    return resp_data, 201



import json
from unittest import TestCase, mock
from unittest.mock import Mock

import flask
import responses
from flask_jwt_extended import create_access_token

from dm.domain.entities import User, ActionTemplate, Server, Software, SoftwareServerAssociation
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.operations import RequestOperation, NativeWaitOperation
from dm.web import create_app, db
from dm.web.network import Response


class TestRequestOperation(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial(user=True, action_template=True)
        self.server = Server.get_current()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dm.use_cases.operations.get')
    @mock.patch('dm.use_cases.operations.create_access_token')
    @mock.patch('dm.use_cases.operations.request')
    def test_execute_send_software(self, mock_request, mock_token, mock_get):
        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000001')
        soft = Software(name='test', version=1, filename='test.zip')
        node1 = Server('node1', port=5000)
        node2 = Server('node2', port=5000)
        ssa1 = SoftwareServerAssociation(software=soft, server=node1, path='/')
        ssa2 = SoftwareServerAssociation(software=soft, server=node2, path='/')
        ssa3 = SoftwareServerAssociation(software=soft, server=self.server, path='/')
        db.session.add_all([soft, node1, node2, ssa1, ssa2, ssa3])

        mock_request.return_value = Response(msg={'transfer_id': 1}, code=at.expected_rc)
        mock_token.return_value = 1
        mock_get.return_value = Response(msg={
            "route_list": [
                {
                    "cost": 0,
                    "destination_id": f"{node1.id}",
                },
                {
                    "cost": 1,
                    "destination_id": f"{self.server.id}",
                }
            ],
        }, code=200, server=node2)

        ro = RequestOperation(at.code,
                              expected_stdout=at.expected_stdout,
                              expected_stderr=at.expected_stderr,
                              expected_rc=at.expected_rc)

        cp = ro.execute(var_context=dict(software_id=str(soft.id), server_id=str(node2.id)), timeout=None)

        mock_request.assert_called_once_with('post', node1, 'api_1_0.send',
                                             json=dict(software_id=str(soft.id), dest_server_id=str(node2.id),
                                                       background=False, include_transfer_data=True, force=True),
                                             auth=HTTPBearerAuth(1))
        self.assertTrue(cp.success)
        self.assertEqual(flask.json.dumps(mock_request.return_value.msg), cp.stdout)

    def test_execute_send_software_no_software(self):
        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000001')

        ro = RequestOperation(at.code,
                              expected_stdout=at.expected_stdout,
                              expected_stderr=at.expected_stderr,
                              expected_rc=at.expected_rc)

        cp = ro.execute(var_context=dict(software_id=1, server_id=str(self.server.id)), timeout=None)

        self.assertFalse(cp.success)
        self.assertEqual(f"software id '1' not found", cp.stderr)

    def test_execute_send_software_no_destination_server(self):
        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000001')
        soft = Software(name='test', version=1, filename='test.zip')
        node1 = Server('node1', port=5000)
        ssa1 = SoftwareServerAssociation(software=soft, server=node1, path='/')
        db.session.add_all([soft, node1, ssa1])

        ro = RequestOperation(at.code,
                              expected_stdout=at.expected_stdout,
                              expected_stderr=at.expected_stderr,
                              expected_rc=at.expected_rc)

        cp = ro.execute(var_context=dict(software_id=soft.id, server_id=str('a')), timeout=None)

        self.assertFalse(cp.success)
        self.assertEqual(f"destination server id 'a' not found", cp.stderr)

    @mock.patch('dm.use_cases.operations.get')
    def test_execute_send_software_no_ssa(self, mock_get):
        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000001')
        soft = Software(name='test', version=1, filename='test.zip')
        node1 = Server('node1', port=5000)
        db.session.add_all([soft, node1])

        mock_get.return_value = Response(code=400)

        ro = RequestOperation(at.code,
                              expected_stdout=at.expected_stdout,
                              expected_stderr=at.expected_stderr,
                              expected_rc=at.expected_rc)

        cp = ro.execute(var_context=dict(software_id=str(soft.id), server_id=str(self.server.id), timeout=None))

        self.assertFalse(cp.success)
        self.assertEqual(f'{soft.id} has no server association', cp.stderr)

    @mock.patch('dm.use_cases.operations.get')
    @mock.patch('dm.use_cases.operations.create_access_token')
    @mock.patch('dm.use_cases.operations.request')
    def test_execute_send_software_error(self, mock_request, mock_token, mock_get):
        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000001')
        soft = Software(name='test', version=1, filename='test.zip')
        node1 = Server('node1', port=5000)
        ssa1 = SoftwareServerAssociation(software=soft, server=node1, path='/')
        db.session.add_all([soft, node1, ssa1])

        mock_request.return_value = Response(msg={'error': 'message'}, code=400)
        mock_token.return_value = 1
        mock_get.return_value = Response(code=400)

        ro = RequestOperation(at.code,
                              expected_stdout=at.expected_stdout,
                              expected_stderr=at.expected_stderr,
                              expected_rc=at.expected_rc)

        cp = ro.execute(var_context=dict(software_id=str(soft.id), server_id=str(node1.id)), timeout=None)

        mock_request.assert_called_once_with('post', node1, 'api_1_0.send',
                                             json=dict(software_id=str(soft.id), dest_server_id=str(node1.id),
                                                       background=False,  include_transfer_data=True, force=True),
                                             auth=HTTPBearerAuth(1))
        self.assertFalse(cp.success)
        self.assertEqual(flask.json.dumps(mock_request.return_value.msg), cp.stdout)

    @responses.activate
    def test_execute_request(self):
        url = 'http://new.url/'
        content = {"content": "this is a message"}
        responses.add(method='GET', url=url, body=json.dumps(content), status=200,
                      content_type='application/json')

        ro = RequestOperation('{"method":"get", "url":"{{view_or_url}}"}',
                              expected_stdout='{}',
                              expected_stderr='',
                              expected_rc=200,
                              post_code="params.update(response.json())")

        params = {"view_or_url": url}
        cp = ro.execute(var_context=params, timeout=None)

        self.assertTrue(cp.success)
        self.assertDictEqual({**params, **content}, params)

    @responses.activate
    def test_execute_request_timeout(self):
        url = 'http://new.url/'
        content = {"content": "this is a message"}
        responses.add(method='GET', url=url, body=TimeoutError())

        ro = RequestOperation('{"method":"get", "url":"{{url}}"}',
                              expected_stdout='{}',
                              expected_stderr='',
                              expected_rc=200
                              )

        params = {"url": url}
        cp = ro.execute(var_context=params, timeout=None)

        self.assertFalse(cp.success)
        self.assertEqual(cp.stderr, "TimeoutError")

    @responses.activate
    def test_execute_request(self):
        url = 'http://new.url/'
        content = "response"
        responses.add(method='GET', url=url, body=content, status=200)

        ro = RequestOperation('{"method":"get", "url":"{{view_or_url}}"}',
                              expected_rc=200)

        params = {"view_or_url": url}
        cp = ro.execute(var_context=params, timeout=None)

        self.assertTrue(cp.success)
        self.assertEqual(content, cp.stdout)


class TestNativeWaitOperation(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial(action_template=True)
        self.server = Server.get_current()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dm.use_cases.operations.db')
    def test_execute(self, mock_db):

        at = ActionTemplate.query.get('00000000-0000-0000-000a-000000000002')

        nwo = NativeWaitOperation(code=at.code, system_kwargs=dict(sleep_time=0.01))

        mock_db.session.query.return_value = Mock()
        mock_db.session.query.return_value.filter_by.return_value = Mock()
        mock_db.session.query.return_value.filter_by.return_value.count.return_value = 0

        cp = nwo.execute(var_context=dict(list_server_names=['node1']), timeout=0.1)

        self.assertEqual(f"Servers node1 not created after 0.1 seconds", cp.stderr)
        self.assertFalse(cp.success)

        cp = nwo.execute(var_context=dict(list_server_names=['node1'], timeout=0.1))
        self.assertEqual(f"Servers node1 not created after 0.1 seconds", cp.stderr)
        self.assertFalse(cp.success)

        cp = nwo.execute(var_context=dict(list_server_names=['node1'], timeout=0.1), timeout=2)
        self.assertEqual(f"Servers node1 not created after 0.1 seconds", cp.stderr)
        self.assertFalse(cp.success)

        mock_db.session.query.return_value.filter_by.return_value.count.side_effect = [0, 0, 1]
        cp = nwo.execute(var_context=dict(list_server_names=['node1']))

        self.assertEqual("Servers node1 found", cp.stdout)
        self.assertIsNone(cp.stderr)
        self.assertTrue(cp.success)

        def func(name):
            if name == 'node1':
                m = Mock()
                m.count.return_value = 1
                return m
            else:
                m = Mock()
                m.count.return_value = 0
                return m

        mock_db.session.query.return_value.filter_by.side_effect = func
        cp = nwo.execute(var_context=dict(list_server_names=['node1', 'node2']), timeout=0.1)
        self.assertEqual(f"Servers node2 not created after 0.1 seconds", cp.stderr)
        self.assertFalse(cp.success)



import itertools

from flask import jsonify
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Server
from dm.web.decorators import securizer, forward_or_dispatch


class GranuleList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        granules = itertools.chain(*[s.granules for s in Server.query.all()])
        return jsonify(list(set(granules)))



import datetime as dt
import re
from unittest import TestCase
from unittest.mock import patch

import responses

from dm.domain.entities import Server, ActionTemplate, ActionType, Catalog, Gate, Route, Locker
from dm.use_cases.use_cases import upgrade_catalog_from_server
from dm.web import create_app, db, errors


class TestUpgradeCatalog(TestCase):

    @patch('dm.domain.entities.get_now')
    def setUp(self, mock_now):
        """Create and configure a new app instance for each test."""
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()

        db.create_all()
        Locker.set_initial()
        Server.set_initial()
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.utils.typos.tzlocal')
    @patch('dm.domain.entities.get_now')
    @patch('dm.use_cases.use_cases.get_distributed_entities')
    @patch('dm.use_cases.use_cases.lock_scope')
    @responses.activate
    def test_upgrade_catalog(self, mock_lock, mock_entities, mock_now, mock_tzlocal):
        mock_lock.return_value.__enter__.return_value = 'applicant'
        mock_entities.return_value = [('ActionTemplate', ActionTemplate), ('Server', Server), ('Gate', Gate)]
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        mock_tzlocal.return_value = dt.timezone.utc

        at1 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='mkdir', version=1,
                             action_type=ActionType.SHELL,
                             code='mkdir {dir}', parameters={}, expected_output=None, expected_rc=None,
                             system_kwargs={})
        db.session.add(at1)
        db.session.commit()

        s = Server(id='aaaaaaaa-1234-5678-1234-56781234bbb1', name='server',
                   last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))
        s_json = s.to_json()
        g = Gate(server=s, port=80, dns='server', last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))
        g_json = g.to_json()
        Route(s, cost=0)

        at2 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa2', name='rmdir', version=1,
                             action_type=ActionType.SHELL,
                             code='rmdir {dir}', parameters={}, expected_output=None, expected_rc=None,
                             system_kwargs={}, last_modified_at=dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc))
        at2_json = at2.to_json()
        del at2

        at1_json = at1.to_json()
        del at1
        at1_json['code'] = 'mkdir -p {dir}'
        responses.add(method='GET',
                      url=re.compile('^' + s.url('api_1_0.catalog', data_mark='12345').replace('12345', '')),
                      json={"ActionTemplate": [at1_json, at2_json], "Server": [s_json], "Gate": [g_json]})

        upgrade_catalog_from_server(s)

        db.session.expire_all()
        atl = [at.to_json() for at in ActionTemplate.query.all()]

        self.assertListEqual([at1_json, at2_json], atl)

        c = Catalog.query.get('ActionTemplate')
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), c.last_modified_at)

        at1 = ActionTemplate.query.get('aaaaaaaa-1234-5678-1234-56781234aaa1')
        self.assertEqual('mkdir -p {dir}', at1.code)

    # no new data
    @patch('dm.use_cases.use_cases.get_distributed_entities')
    @patch('dm.use_cases.use_cases.lock_scope')
    @responses.activate
    def test_upgrade_catalog_no_data(self, mock_lock, mock_entities):
        mock_lock.return_value.__enter__.return_value = 'applicant'
        mock_entities.return_value = [('ActionTemplate', ActionTemplate)]

        s = Server('server', last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))
        g = Gate(server=s, port=80)
        Route(s, cost=0)

        responses.add(method='GET',
                      url=re.compile('^' + s.url('api_1_0.catalog', data_mark='12345').replace('12345', '')),
                      json={"ActionTemplate": []})

        upgrade_catalog_from_server(s)

        atl = [at.to_json() for at in ActionTemplate.query.all()]

        self.assertEqual(0, len(atl))

    @patch('dm.use_cases.use_cases.get_distributed_entities')
    @patch('dm.use_cases.use_cases.lock_scope')
    @responses.activate
    def test_upgrade_catalog_catalog_mismatch(self, mock_lock, mock_entities):
        mock_lock.return_value.__enter__.return_value = 'applicant'
        mock_entities.return_value = [('ActionTemplate', ActionTemplate), ('Server', Server)]

        s = Server('server', last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), port=8000)
        Route(s, cost=0)

        at1 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='mkdir', version=1,
                             action_type=ActionType.SHELL,
                             code='mkdir {dir}', parameters={}, expected_output=None, expected_rc=None,
                             system_kwargs={}, last_modified_at=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc))

        at2 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa2', name='rmdir', version=1,
                             action_type=ActionType.SHELL,
                             code='rmdir {dir}', parameters={}, expected_output=None, expected_rc=None,
                             system_kwargs={}, last_modified_at=dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc))

        responses.add(method='GET',
                      url=re.compile('^' + s.url('api_1_0.catalog', data_mark='12345').replace('12345', '')),
                      json={"ActionTemplate": [at1.to_json(), at2.to_json()]})

        with self.assertRaises(errors.CatalogMismatch):
            upgrade_catalog_from_server(s)



import os

from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Software, Server, User
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import md5
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class TestSoftwareList(TestCaseLockBypass):

    def setUp(self) -> None:
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))
        soft1 = Software(id='11111111-2222-3333-4444-555555550001', name='Dimensigon', version='0.0.1',
                         filename='Dimensigon_0.0.1.tar.gz')
        soft2 = Software(id='11111111-2222-3333-4444-555555550002', name='Dimensigon', version='0.0.2',
                         filename='Dimensigon_0.0.2.tar.gz')
        soft3 = Software(id='11111111-2222-3333-4444-555555550003', name='python', version='3.6.8',
                         filename='python_3.6.8.x64.tar.gz')
        db.session.add_all([soft1, soft2, soft3])
        db.session.commit()
        self.soft1_json = soft1.to_json()
        self.soft2_json = soft2.to_json()
        self.soft3_json = soft3.to_json()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.softwarelist'), headers=self.auth.header)

        self.assertListEqual(
            [self.soft1_json, self.soft2_json, self.soft3_json], resp.get_json())

    def test_get_with_filter(self):
        resp = self.client.get(url_for('api_1_0.softwarelist', **{'filter[name]': 'Dimensigon'}),
                               headers=self.auth.header)

        self.assertListEqual([self.soft1_json, self.soft2_json], resp.get_json())

    def test_get_with_filter2(self):
        resp = self.client.get(url_for('api_1_0.softwarelist', **{'filter[version]': '0.0.1,3.6.8'}),
                               headers=self.auth.header)

        self.assertListEqual([self.soft1_json, self.soft3_json], resp.get_json())

    def test_post(self):
        size = os.path.getsize(__file__)
        checksum = md5(__file__)
        filename = os.path.basename(__file__)
        data = dict(name="Dimensigon", version="0.0.3", family='middleware', server_id=str(Server.get_current().id),
                    file=__file__)
        resp = self.client.post(url_for('api_1_0.softwarelist'), headers=self.auth.header, json=data)

        self.assertEqual(201, resp.status_code)

        soft = Software.query.filter_by(name="Dimensigon", version="0.0.3").one()
        self.assertEqual(size, soft.size)
        self.assertEqual(checksum, soft.checksum)
        self.assertEqual(filename, soft.filename)
        self.assertEqual(1, len(soft.ssas))
        ssa = soft.ssas[0]

        self.assertEqual(os.path.dirname(__file__), ssa.path)



import datetime
import re
import sys
import threading
import traceback
import typing as t
from contextlib import contextmanager
from json import JSONEncoder

from flask import current_app, request
from flask_sqlalchemy import BaseQuery
from sqlalchemy.orm import sessionmaker

from dm import defaults
from dm.utils.asyncio import run
from dm.utils.helpers import is_iterable_not_string
from dm.utils.typos import Id
from dm.web.errors import EntityNotFound, NoDataFound

if t.TYPE_CHECKING:
    from dm.domain.entities import Server


class BaseQueryJSON(BaseQuery):
    """SQLAlchemy :class:`~sqlalchemy.orm.query.Query` subclass with convenience methods for querying in a web application.

    This is the default :attr:`~Model.query` object used for models, and exposed as :attr:`~SQLAlchemy.Query`.
    Override the query class for an individual model by subclassing this and setting :attr:`~Model.query_class`.
    """

    def get_or_404(self, ident, description=None):
        """Like :meth:`get` but aborts with 404 if not found instead of returning ``None``."""

        rv = self.get(ident)
        if rv is None:
            # abort(format_error_response(EntityNotFound(self.column_descriptions[0]['name'], ident)))
            raise EntityNotFound(self.column_descriptions[0]['name'], ident)
        return rv

    def first_or_404(self, description=None):
        """Like :meth:`first` but aborts with 404 if not found instead of returning ``None``."""

        rv = self.first()
        if rv is None:
            # abort(format_error_response(NoDataFound(self.column_descriptions[0]['name'])))
            raise NoDataFound(self.column_descriptions[0]['name'])
        return rv

class QueryWithSoftDelete(BaseQueryJSON):
    _with_deleted = False

    def __new__(cls, *args, **kwargs):
        obj = super(QueryWithSoftDelete, cls).__new__(cls)
        obj._with_deleted = kwargs.pop('_with_deleted', False)
        if len(args) > 0:
            super(QueryWithSoftDelete, obj).__init__(*args, **kwargs)
            return obj.filter_by(deleted=False) if not obj._with_deleted else obj
        return obj

    def __init__(self, *args, **kwargs):
        pass

    def with_deleted(self):
        from dm.web import db
        return self.__class__(db.class_mapper(self._mapper_zero().class_),
                              session=db.session(), _with_deleted=True)

    def _get(self, *args, **kwargs):
        # this calls the original query.get function from the base class
        return super(QueryWithSoftDelete, self).get(*args, **kwargs)

    def get(self, *args, **kwargs):
        # the query.get method does not like it if there is a filter clause
        # pre-loaded, so we need to implement it using a workaround
        obj = self.with_deleted()._get(*args, **kwargs)
        return obj if obj is None or self._with_deleted or not obj.deleted else None


def filter_query(entity, filters, exclude: t.Container = None):
    """Generates a sqlalchemy query object filtered by filters.

    entity: entity to filter
    filters: filters in JSON API format https://jsonapi.org/format/#fetching-filtering
    exclude: columns to exclude on filter

    """
    filters = [(re.search('^filter\[(\w+)\]$', k).group(1), v) for k, v in filters.items() if
               k.startswith('filter[')]
    query = entity.query
    for k, v in filters:
        column = getattr(entity, k, None)
        if not column or k in (exclude or []):
            return {'error': f'Invalid filter column: {k}'}, 404
        if ',' in v:
            values = v.split(',')
            query = query.filter(column.in_(values))
        else:
            if v.lower() == 'true':
                v = True
            elif v.lower() == 'false':
                v = False
            query = query.filter(column == v)
    return query


def check_param_in_uri(param):
    params = request.args.get('params', '')
    params = params.split(',')
    return param in params


def run_in_background(func: t.Callable, app=None, args=None, kwargs=None):
    args = args or ()
    kwargs = kwargs or {}
    try:
        app = current_app._get_current_object()
    except RuntimeError:
        app = app

    def thread_with_app_context():
        with app.app_context():
            run(func(*args, **kwargs))

    th = threading.Thread(target=thread_with_app_context)
    # th.daemon = True
    th.start()


class DatetimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime(defaults.DATETIME_FORMAT)
        return JSONEncoder.default(self, obj)


def json_format_error():
    return {'error': format_error()}


def format_error():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    return traceback.format_exc() if current_app.config['DEBUG'] else str(exc_value)


def search(server_or_granule, servers: t.List['Server'], ids: bool = False) -> t.Union[t.List['Server'], t.List[Id]]:
    """
    Searches the server from list of servers
    Args:
        server_or_granule: server id, name or granule to search
        servers: list of servers
        ids: determines if it should return a list of ids or Servers

    Returns:

    """
    if server_or_granule == 'all':
        return servers
    server_list = [(str(server.id) if ids else server) for server in servers if server.id == server_or_granule]
    if not server_list:
        server_list = [(str(server.id) if ids else server) for server in servers if server_or_granule == server.name]
        if not server_list:
            server_list = [(str(server.id) if ids else server) for server in servers if
                           server_or_granule in server.granules]
    return server_list


def normalize_hosts(hosts: t.Dict[str, t.Union[str, t.List[str]]]) -> t.List[str]:
    """ normalizes all str to server ids

    Args:
        hosts: data structure containing name, granules or id to convert into

    Returns:
        list: servers not found in catalog

    Examples:
        >>> normalize_hosts({'front': 'node1', 'back': 'tibero', 'all': ['node1', 'tibero', 'id_node4']})
        { 'all': ['id_node1', 'id_tibero_node2', 'id_tibero_node3', 'id_node4'],
          'databases': ['id_tibero_node2', 'id_tibero_node3'],
          'front': ['id_node1']}

    """
    from dm.domain.entities import Server

    not_found = []
    servers = Server.query.all()
    for target, v in hosts.items():
        server_list = []
        if is_iterable_not_string(v):
            for vv in v:
                sl = search(vv, servers, ids=True)
                if len(sl) == 0:
                    not_found.append(vv)
                else:
                    server_list.extend(sl)
        else:
            sl = search(v, servers, ids=True)
            if len(sl) == 0:
                not_found.append(v)
            else:
                server_list.extend(sl)
        hosts[target] = server_list
    return not_found


@contextmanager
def session():
    from dm.web import db
    engine = db.get_engine()
    Session = sessionmaker(bind=engine)
    s = Session()
    try:
        yield s
    finally:
        s.close()
